import logging

import datadoc_model.all_optional.model as all_optional_model
import pytest
from pydantic import AnyUrl
from typeguard import suppress_type_checks

from dapla_metadata.datasets.core import Datadoc
from dapla_metadata.datasets.utility.urn import URN_ERROR_MESSAGE_BASE
from dapla_metadata.datasets.utility.urn import ReferenceUrlTypes
from dapla_metadata.datasets.utility.urn import SsbNaisDomains
from dapla_metadata.datasets.utility.urn import UrlVisibility
from dapla_metadata.datasets.utility.urn import klass_urn_converter
from dapla_metadata.datasets.utility.urn import vardef_urn_converter

EXAMPLE_KLASS_ID = "91"
EXAMPLE_KLASS_URN = AnyUrl(klass_urn_converter.get_urn(EXAMPLE_KLASS_ID))
EXAMPLE_VARDEF_ID = "hd8_Ks-9"
EXAMPLE_VARDEF_URN = AnyUrl(vardef_urn_converter.get_urn(EXAMPLE_VARDEF_ID))

VARIABLE_DEFINITION_URN_TEST_CASES = [
    *[
        (
            f"https://metadata.{domain.value}/variable-definitions/{EXAMPLE_VARDEF_ID}",
            EXAMPLE_VARDEF_URN,
            False,
        )
        for domain in SsbNaisDomains
    ],
    *[
        (
            f"https://catalog.{domain.value}/variable-definitions/{EXAMPLE_VARDEF_ID}",
            EXAMPLE_VARDEF_URN,
            False,
        )
        for domain in SsbNaisDomains
    ],
    (
        EXAMPLE_VARDEF_URN,
        EXAMPLE_VARDEF_URN,
        False,
    ),
]


@pytest.mark.parametrize(
    ("case", "expected_result", "expect_warning"),
    [
        *VARIABLE_DEFINITION_URN_TEST_CASES,
        ("https://www.vg.no", None, True),
    ],
)
def test_convert_to_vardef_urn(
    case: str,
    expected_result: AnyUrl | None,
    expect_warning: bool,  # noqa: ARG001
):
    assert vardef_urn_converter.convert_url_to_urn(case) == expected_result


@pytest.mark.parametrize(
    ("case", "expected_result", "expect_warning"),
    [
        *VARIABLE_DEFINITION_URN_TEST_CASES,
        ("https://www.vg.no", AnyUrl("https://www.vg.no/"), True),
        (None, None, False),
    ],
)
def test_convert_to_vardef_urn_end_to_end(
    case: str | None,
    expected_result: str | None,
    expect_warning: bool,
    caplog: pytest.LogCaptureFixture,
):
    caplog.set_level(logging.ERROR)
    meta = Datadoc()
    meta._set_metadata(  # noqa: SLF001
        all_optional_model.DatadocMetadata(
            dataset=all_optional_model.Dataset(),
            variables=[
                all_optional_model.Variable(
                    short_name="my_variable", definition_uri=case
                )
            ],
        )
    )
    if not expected_result:
        assert meta.variables[0].definition_uri is None
    else:
        assert meta.variables[0].definition_uri == expected_result
    assert (URN_ERROR_MESSAGE_BASE in caplog.text) is expect_warning


@pytest.mark.parametrize("identifier", [EXAMPLE_VARDEF_ID])
@pytest.mark.parametrize("visibility", ["internal", "public"])
@pytest.mark.parametrize(
    "url_type", [ReferenceUrlTypes.API, ReferenceUrlTypes.FRONTEND]
)
def test_vardef_get_url(
    url_type: ReferenceUrlTypes, visibility: UrlVisibility, identifier: str
):
    url = vardef_urn_converter.get_url(identifier, url_type, visibility)
    assert url is not None
    assert url.endswith("/" + identifier)
    if url_type == ReferenceUrlTypes.API:
        assert url.startswith("https://metadata.")
    else:
        assert url.startswith("https://catalog.")
    if visibility == "internal":
        assert ".intern." in url
    else:
        assert ".intern." not in url


@pytest.mark.parametrize(
    ("urn_or_url", "identifier"),
    [
        ("https://www.vg.no", None),
        (EXAMPLE_VARDEF_URN, EXAMPLE_VARDEF_ID),
        (
            f"https://catalog.ssb.no/variable-definitions/{EXAMPLE_VARDEF_ID}",
            EXAMPLE_VARDEF_ID,
        ),
        (
            AnyUrl(
                f"https://metadata.intern.test.ssb.no/variable-definitions/{EXAMPLE_VARDEF_ID}"
            ),
            EXAMPLE_VARDEF_ID,
        ),
    ],
)
def test_vardef_get_id(urn_or_url: str | AnyUrl, identifier: str | None):
    assert vardef_urn_converter.get_id(urn_or_url) == identifier


@pytest.mark.parametrize(
    ("identifier", "expected"),
    [
        (EXAMPLE_VARDEF_ID, True),
        ("invalid-id", False),
        ("123", False),
        ("12345678", True),
        ("123456789", False),
        ("abc$EFGH", False),
        ("", False),
        (None, False),
    ],
)
@suppress_type_checks
def test_vardef_is_id(identifier: str, expected: bool):
    assert vardef_urn_converter.is_id(identifier) is expected


CLASSIFICATION_URN_TEST_CASES = [
    (
        f"https://www.ssb.no/klass/klassifikasjoner/{EXAMPLE_KLASS_ID}",
        EXAMPLE_KLASS_URN,
        False,
    ),
    (
        f"https://www.ssb.no/en/klass/klassifikasjoner/{EXAMPLE_KLASS_ID}",
        EXAMPLE_KLASS_URN,
        False,
    ),
    (
        f"https://data.ssb.no/api/klass/v1/classifications/{EXAMPLE_KLASS_ID}",
        EXAMPLE_KLASS_URN,
        False,
    ),
    (
        f"https://data.ssb.no/api/klass/v1/classifications/{EXAMPLE_KLASS_ID}.json",
        EXAMPLE_KLASS_URN,
        False,
    ),
    (
        EXAMPLE_KLASS_URN,
        EXAMPLE_KLASS_URN,
        False,
    ),
]


@pytest.mark.parametrize(
    ("case", "expected_result", "expect_warning"),
    CLASSIFICATION_URN_TEST_CASES,
)
def test_convert_to_klass_urn(
    case: str,
    expected_result: str,
    expect_warning: bool,  # noqa: ARG001
):
    assert klass_urn_converter.convert_url_to_urn(case) == expected_result


@pytest.mark.parametrize(
    ("case", "expected_result", "expect_warning"),
    [
        *CLASSIFICATION_URN_TEST_CASES,
        ("https://www.vg.no", AnyUrl("https://www.vg.no/"), True),
        (None, None, False),
    ],
)
def test_convert_to_klass_urn_end_to_end(
    case: str | None,
    expected_result: str | None,
    expect_warning: bool,
    caplog: pytest.LogCaptureFixture,
):
    caplog.set_level(logging.ERROR)
    meta = Datadoc()
    meta._set_metadata(  # noqa: SLF001
        all_optional_model.DatadocMetadata(
            dataset=all_optional_model.Dataset(),
            variables=[
                all_optional_model.Variable(
                    short_name="my_variable", classification_uri=case
                )
            ],
        )
    )
    if not expected_result:
        assert meta.variables[0].classification_uri is None
    else:
        assert meta.variables[0].classification_uri == expected_result
    assert (URN_ERROR_MESSAGE_BASE in caplog.text) is expect_warning


@pytest.mark.parametrize("identifier", [EXAMPLE_KLASS_ID])
@pytest.mark.parametrize("visibility", ["internal", "public"])
@pytest.mark.parametrize(
    "url_type", [ReferenceUrlTypes.API, ReferenceUrlTypes.FRONTEND]
)
def test_klass_get_url(
    url_type: ReferenceUrlTypes, visibility: UrlVisibility, identifier: str
):
    url = klass_urn_converter.get_url(identifier, url_type, visibility)
    if visibility == "internal":
        # Klass doesn't have an internal representation
        assert url is None
    else:
        assert url is not None
        assert url.endswith("/" + identifier)
        if url_type == ReferenceUrlTypes.API:
            assert url.startswith("https://data.")
        else:
            assert url.startswith("https://www.ssb.")


@pytest.mark.parametrize(
    ("urn_or_url", "identifier"),
    [
        ("https://www.vg.no", None),
        (EXAMPLE_KLASS_URN, EXAMPLE_KLASS_ID),
        (
            f"https://www.ssb.no/klass/klassifikasjoner/{EXAMPLE_KLASS_ID}",
            EXAMPLE_KLASS_ID,
        ),
        (
            AnyUrl(
                f"https://data.ssb.no/api/klass/v1/classifications/{EXAMPLE_KLASS_ID}"
            ),
            EXAMPLE_KLASS_ID,
        ),
    ],
)
def test_klass_get_id(urn_or_url: str | AnyUrl, identifier: str | None):
    assert klass_urn_converter.get_id(urn_or_url) == identifier


@pytest.mark.parametrize(
    ("identifier", "expected"),
    [
        (EXAMPLE_KLASS_ID, True),
        ("invalid-id", False),
        ("123", True),
        ("1", True),
        ("123456789", False),
        ("abcdEFGH", False),
        ("", False),
        (None, False),
    ],
)
@suppress_type_checks
def test_klass_is_id(identifier: str, expected: bool):
    assert klass_urn_converter.is_id(identifier) is expected



import logging
from pathlib import Path

import pytest
from pytz import UnknownTimeZoneError
from ruamel.yaml import YAMLError

from dapla_metadata.variable_definitions._utils.config import WORKSPACE_DIR
from dapla_metadata.variable_definitions.exceptions import VardefFileError
from dapla_metadata.variable_definitions.vardef import Vardef

LOGGER = logging.getLogger(__name__)


@pytest.mark.usefixtures("set_temp_workspace")
def test_write_template():
    file_path = Vardef.write_template_to_file()
    assert file_path.exists()


def test_write_template_no_workspace(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.delenv(WORKSPACE_DIR, raising=False)
    with pytest.raises(
        VardefFileError,
        match="WORKSPACE_DIR is not set",
    ):
        Vardef.write_template_to_file()


def test_write_template_no_work_folder(monkeypatch: pytest.MonkeyPatch):
    """'WORKSPACE_DIR' is set but there is no path matching value."""
    monkeypatch.setenv(WORKSPACE_DIR, "statistics/a/work")

    with pytest.raises(
        FileNotFoundError,
    ):
        Vardef.write_template_to_file()


@pytest.mark.usefixtures("set_temp_workspace")
@pytest.mark.parametrize(
    "file_path",
    ["statistics/a/work", "./", "../..", "../", "./work"],
)
def test_write_template_workspace(tmp_path: Path, file_path, monkeypatch):
    """Assert file is created at the correct path no matter starting point filesystem."""
    base_path = (tmp_path / file_path).resolve()
    base_path.mkdir(parents=True, exist_ok=True)

    # Mock current directory using monkeypatch
    monkeypatch.setattr(Path, "cwd", lambda: base_path)

    file_path = Vardef.write_template_to_file()

    assert file_path.exists()


@pytest.mark.usefixtures("set_temp_workspace")
@pytest.mark.parametrize(
    "custom_path",
    ["serious", "statistics/april", "private", "toppers$#23"],
)
def test_write_template_to_custom_path_directory_doesnt_exist(
    tmp_path: Path,
    custom_path: str,
):
    file_path = Vardef.write_template_to_file(custom_file_path=tmp_path / custom_path)

    assert file_path.exists()

    file_path_str = str(file_path)

    # Not saved at default path
    assert "work" not in file_path_str
    assert "variable-definitions" not in file_path_str
    # Saved at custom path
    assert custom_path in file_path_str

    # Filename is correct
    file_name = file_path.name
    file_name_minus_timestamp = file_name.rsplit("_", 1)[0] + ".yaml"
    assert file_name_minus_timestamp == "variable_definition_template.yaml"


@pytest.mark.usefixtures("set_temp_workspace")
@pytest.mark.parametrize(
    "custom_path",
    ["problems", "statistics", "private", "toppers$#23"],
)
def test_write_template_to_custom_path_directory_exist(
    tmp_path: Path,
    custom_path: str,
):
    local_dir = tmp_path / custom_path
    local_dir.mkdir(parents=True, exist_ok=True)
    file_path = Vardef.write_template_to_file(custom_file_path=local_dir)

    assert file_path.exists()

    file_path_str = str(file_path)

    # Not saved at default path
    assert "work" not in file_path_str
    assert "variable-definitions" not in file_path_str
    # Saved at custom path
    assert custom_path in file_path_str

    # Filename is correct
    file_name = file_path.name
    file_name_minus_timestamp = file_name.rsplit("_", 1)[0] + ".yaml"
    assert file_name_minus_timestamp == "variable_definition_template.yaml"


@pytest.mark.usefixtures("_delete_workspace_dir_env_var")
def test_write_template_to_custom_path_no_workspace_dir_env(tmp_path: Path):
    file_path = Vardef.write_template_to_file(custom_file_path=tmp_path / "cki/job")
    assert file_path.exists()


# check permission error
@pytest.mark.usefixtures("set_temp_workspace")
@pytest.mark.parametrize(
    ("custom_file_path", "expected_error"),
    [
        ("my_file.yaml", ValueError),
        ("\0", ValueError),
        ("a" * 300, OSError),
    ],
)
def test_write_template_to_invalid_custom_directory(
    custom_file_path,
    expected_error,
):
    with pytest.raises(expected_error):
        Vardef.write_template_to_file(custom_file_path=custom_file_path)


@pytest.mark.parametrize(
    ("mock_target", "side_effect"),
    [
        (
            "dapla_metadata.variable_definitions._utils.template_files._model_to_yaml_with_comments",
            FileExistsError,
        ),
        (
            "dapla_metadata.variable_definitions._utils.template_files._model_to_yaml_with_comments",
            PermissionError,
        ),
        (
            "dapla_metadata.variable_definitions._utils.template_files._get_current_time",
            UnknownTimeZoneError,
        ),
        (
            "dapla_metadata.variable_definitions._utils.template_files._model_to_yaml_with_comments",
            YAMLError,
        ),
        (
            "dapla_metadata.variable_definitions._utils.template_files._model_to_yaml_with_comments",
            EOFError,
        ),
        (
            "dapla_metadata.variable_definitions._utils.files._get_workspace_dir",
            NotADirectoryError,
        ),
    ],
)
def test_write_template_exceptions(mock_target: str, side_effect: Exception, mocker):
    """Test that write_template_to_file raises VardefFileError for different exceptions."""
    mocker.patch(mock_target, side_effect=side_effect)

    with pytest.raises(VardefFileError):
        Vardef.write_template_to_file()


def test_logging_workspace_dir(caplog):
    """Test logging intended for user."""
    caplog.set_level(logging.INFO)
    Vardef.write_template_to_file()
    assert "Created editable variable definition template file" in caplog.text



"""Tests for pseudonymization."""

from __future__ import annotations

import shutil
from dataclasses import dataclass
from datetime import datetime
from datetime import timezone
from typing import TYPE_CHECKING

import pytest
from datadoc_model.all_optional.model import (
    Pseudonymization as PseudonymizationOptional,
)
from datadoc_model.required.model import Pseudonymization as PseudonymizationRequired

from dapla_metadata.datasets.core import Datadoc
from dapla_metadata.datasets.utility.constants import DAEAD_ENCRYPTION_KEY_REFERENCE
from dapla_metadata.datasets.utility.constants import ENCRYPTION_PARAMETER_KEY_ID
from dapla_metadata.datasets.utility.constants import ENCRYPTION_PARAMETER_SNAPSHOT_DATE
from dapla_metadata.datasets.utility.constants import ENCRYPTION_PARAMETER_STRATEGY
from dapla_metadata.datasets.utility.constants import ENCRYPTION_PARAMETER_STRATEGY_SKIP
from dapla_metadata.datasets.utility.constants import PAPIS_ENCRYPTION_KEY_REFERENCE
from dapla_metadata.datasets.utility.constants import PAPIS_STABLE_IDENTIFIER_TYPE
from dapla_metadata.datasets.utility.enums import EncryptionAlgorithm
from dapla_metadata.datasets.utility.utils import PseudonymizationType
from dapla_metadata.datasets.utility.utils import get_current_date
from tests.datasets.constants import TEST_EXISTING_METADATA_FILE_NAME
from tests.datasets.constants import TEST_EXISTING_METADATA_NAMING_STANDARD_FILEPATH
from tests.datasets.constants import TEST_PARQUET_FILE_NAME
from tests.datasets.constants import TEST_PARQUET_FILEPATH
from tests.datasets.constants import TEST_PSEUDO_DIRECTORY

if TYPE_CHECKING:
    from datetime import datetime  # noqa: TC004
    from pathlib import Path

    from upath import UPath

    from dapla_metadata.datasets.statistic_subject_mapping import (
        StatisticSubjectMapping,
    )


@dataclass
class PseudoCase:
    new_pseudo: PseudonymizationOptional | None = None
    expected_algorithm: str | None = None
    expected_stable_type: str | None = None
    expected_key: str | None = None
    expected_params: list[dict] | None = None
    expected_pseudo_time: datetime | None = None
    expected_stable_version: str | None = None


def _assert_dicts_in_list(expected: list[dict] | None, actual: list[dict]):
    if expected:
        for d in expected:
            assert d in actual, f"Missing expected dict: {d}"


@pytest.mark.parametrize(
    "case",
    [
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
            ),
            expected_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
            expected_key=PAPIS_ENCRYPTION_KEY_REFERENCE,
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: PAPIS_ENCRYPTION_KEY_REFERENCE},
                {ENCRYPTION_PARAMETER_STRATEGY: ENCRYPTION_PARAMETER_STRATEGY_SKIP},
            ],
        ),
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
                encryption_key_reference="fall-out-2",
            ),
            expected_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
            expected_key="fall-out-2",
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: PAPIS_ENCRYPTION_KEY_REFERENCE},
                {ENCRYPTION_PARAMETER_STRATEGY: ENCRYPTION_PARAMETER_STRATEGY_SKIP},
            ],
        ),
        PseudoCase(
            new_pseudo=None,
            expected_algorithm=None,
        ),
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
                pseudonymization_time=datetime(
                    2018, 3, 3, 12, 30, 0, tzinfo=timezone.utc
                ),
            ),
            expected_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
            expected_key=PAPIS_ENCRYPTION_KEY_REFERENCE,
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: PAPIS_ENCRYPTION_KEY_REFERENCE},
                {ENCRYPTION_PARAMETER_STRATEGY: ENCRYPTION_PARAMETER_STRATEGY_SKIP},
            ],
            expected_pseudo_time=datetime(2018, 3, 3, 12, 30, 0, tzinfo=timezone.utc),
        ),
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
                encryption_algorithm_parameters=[{"someKey": "someValue"}],
            ),
            expected_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
            expected_key=PAPIS_ENCRYPTION_KEY_REFERENCE,
            expected_params=[
                {"someKey": "someValue"},
                {ENCRYPTION_PARAMETER_KEY_ID: PAPIS_ENCRYPTION_KEY_REFERENCE},
                {ENCRYPTION_PARAMETER_STRATEGY: ENCRYPTION_PARAMETER_STRATEGY_SKIP},
            ],
        ),
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
                encryption_algorithm_parameters=[
                    {ENCRYPTION_PARAMETER_KEY_ID: "heaven-ref-3"}
                ],
            ),
            expected_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
            expected_key=PAPIS_ENCRYPTION_KEY_REFERENCE,
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: "heaven-ref-3"},
                {ENCRYPTION_PARAMETER_STRATEGY: ENCRYPTION_PARAMETER_STRATEGY_SKIP},
            ],
        ),
    ],
    ids=[
        "Add new pseudonymization PAPIS without stable ID",
        "Add new pseudonymization PAPIS without stable ID - encryption key refrence set",
        "No pseudonymization - empty pseudonymization",
        "Add new pseudonymization PAPIS without stable ID - pseudonymization time",
        "Add new pseudonymization PAPIS without stable ID - additional algorithm parameters",
        "Add new pseudonymization PAPIS without stable ID - keyId in algorithm parameters",
    ],
)
def test_add_default_pseudonymization_values_papis_without_stable_id(
    case: PseudoCase, metadata: Datadoc
):
    sykepenger = metadata.variables_lookup["sykepenger"]

    assert sykepenger.short_name is not None

    metadata.add_pseudonymization(sykepenger.short_name, case.new_pseudo)
    assert sykepenger.pseudonymization is not None
    if sykepenger.pseudonymization:
        assert (
            sykepenger.pseudonymization.encryption_algorithm == case.expected_algorithm
        )
        assert sykepenger.pseudonymization.encryption_key_reference == case.expected_key
        assert (
            sykepenger.pseudonymization.pseudonymization_time
            == case.expected_pseudo_time
        )
        if sykepenger.pseudonymization.encryption_algorithm_parameters is None:
            assert case.expected_params is None
        else:
            _assert_dicts_in_list(
                case.expected_params,
                sykepenger.pseudonymization.encryption_algorithm_parameters,
            )


@pytest.mark.parametrize(
    "case",
    [
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
                stable_identifier_type=PAPIS_STABLE_IDENTIFIER_TYPE,
            ),
            expected_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
            expected_stable_type=PAPIS_STABLE_IDENTIFIER_TYPE,
            expected_key=PAPIS_ENCRYPTION_KEY_REFERENCE,
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: PAPIS_ENCRYPTION_KEY_REFERENCE},
                {ENCRYPTION_PARAMETER_SNAPSHOT_DATE: get_current_date()},
                {ENCRYPTION_PARAMETER_STRATEGY: ENCRYPTION_PARAMETER_STRATEGY_SKIP},
            ],
        ),
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
                stable_identifier_type=PAPIS_STABLE_IDENTIFIER_TYPE,
                encryption_key_reference="oh-no-4",
            ),
            expected_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
            expected_stable_type=PAPIS_STABLE_IDENTIFIER_TYPE,
            expected_key="oh-no-4",
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: PAPIS_ENCRYPTION_KEY_REFERENCE},
                {ENCRYPTION_PARAMETER_SNAPSHOT_DATE: get_current_date()},
                {ENCRYPTION_PARAMETER_STRATEGY: ENCRYPTION_PARAMETER_STRATEGY_SKIP},
            ],
        ),
        PseudoCase(
            new_pseudo=None,
        ),
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
                stable_identifier_type=PAPIS_STABLE_IDENTIFIER_TYPE,
                pseudonymization_time=datetime(
                    2018, 3, 3, 12, 30, 0, tzinfo=timezone.utc
                ),
            ),
            expected_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
            expected_stable_type=PAPIS_STABLE_IDENTIFIER_TYPE,
            expected_key=PAPIS_ENCRYPTION_KEY_REFERENCE,
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: PAPIS_ENCRYPTION_KEY_REFERENCE},
                {ENCRYPTION_PARAMETER_SNAPSHOT_DATE: get_current_date()},
                {ENCRYPTION_PARAMETER_STRATEGY: ENCRYPTION_PARAMETER_STRATEGY_SKIP},
            ],
            expected_pseudo_time=datetime(2018, 3, 3, 12, 30, 0, tzinfo=timezone.utc),
        ),
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
                stable_identifier_type=PAPIS_STABLE_IDENTIFIER_TYPE,
                stable_identifier_version="3012-10-11",
            ),
            expected_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
            expected_stable_type=PAPIS_STABLE_IDENTIFIER_TYPE,
            expected_key=PAPIS_ENCRYPTION_KEY_REFERENCE,
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: PAPIS_ENCRYPTION_KEY_REFERENCE},
                {ENCRYPTION_PARAMETER_SNAPSHOT_DATE: get_current_date()},
                {ENCRYPTION_PARAMETER_STRATEGY: ENCRYPTION_PARAMETER_STRATEGY_SKIP},
            ],
            expected_stable_version="3012-10-11",
        ),
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
                stable_identifier_type=PAPIS_STABLE_IDENTIFIER_TYPE,
                encryption_algorithm_parameters=[
                    {ENCRYPTION_PARAMETER_SNAPSHOT_DATE: "2025-12-12"}
                ],
            ),
            expected_algorithm=EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value,
            expected_stable_type=PAPIS_STABLE_IDENTIFIER_TYPE,
            expected_key=PAPIS_ENCRYPTION_KEY_REFERENCE,
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: PAPIS_ENCRYPTION_KEY_REFERENCE},
                {ENCRYPTION_PARAMETER_STRATEGY: ENCRYPTION_PARAMETER_STRATEGY_SKIP},
            ],
        ),
    ],
    ids=[
        "Add new pseudonymization PAPIS with stable ID",
        "Add pseudonymization PAPIS with stable ID - encryption key reference set",
        "No pseudonymization - empty pseudonymization",
        "Add pseudonymization time PAPIS with stable ID - pseudonymization time",
        "Add pseudonymization time PAPIS with stable ID - stable version",
        "Add pseudonymization time PAPIS with stable ID - snapshotDate set",
    ],
)
def test_add_default_pseudonymization_values_papis_with_stable_id(
    case: PseudoCase, metadata: Datadoc
):
    sykepenger = metadata.variables_lookup["sykepenger"]

    assert sykepenger.short_name is not None

    metadata.add_pseudonymization(sykepenger.short_name, case.new_pseudo)
    assert sykepenger.pseudonymization is not None
    assert (
        sykepenger.pseudonymization.stable_identifier_type == case.expected_stable_type
    )
    assert sykepenger.pseudonymization.encryption_algorithm == case.expected_algorithm
    assert sykepenger.pseudonymization.encryption_key_reference == case.expected_key
    assert (
        sykepenger.pseudonymization.stable_identifier_version
        == case.expected_stable_version
    )
    assert (
        sykepenger.pseudonymization.pseudonymization_time == case.expected_pseudo_time
    )
    algorithm_params = sykepenger.pseudonymization.encryption_algorithm_parameters
    if algorithm_params is None:
        assert case.expected_params is None
    else:
        _assert_dicts_in_list(
            case.expected_params,
            algorithm_params,
        )


@pytest.mark.parametrize(
    "case",
    [
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.DAEAD_ENCRYPTION_ALGORITHM.value,
            ),
            expected_algorithm=EncryptionAlgorithm.DAEAD_ENCRYPTION_ALGORITHM.value,
            expected_key=DAEAD_ENCRYPTION_KEY_REFERENCE,
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: DAEAD_ENCRYPTION_KEY_REFERENCE},
            ],
        ),
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.DAEAD_ENCRYPTION_ALGORITHM.value,
                encryption_key_reference="jippi-ya",
            ),
            expected_algorithm=EncryptionAlgorithm.DAEAD_ENCRYPTION_ALGORITHM.value,
            expected_key="jippi-ya",
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: DAEAD_ENCRYPTION_KEY_REFERENCE},
            ],
        ),
        PseudoCase(
            new_pseudo=None,
            expected_algorithm=None,
        ),
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.DAEAD_ENCRYPTION_ALGORITHM.value,
                pseudonymization_time=datetime(
                    2018, 3, 3, 12, 30, 0, tzinfo=timezone.utc
                ),
            ),
            expected_algorithm=EncryptionAlgorithm.DAEAD_ENCRYPTION_ALGORITHM.value,
            expected_key=DAEAD_ENCRYPTION_KEY_REFERENCE,
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: DAEAD_ENCRYPTION_KEY_REFERENCE},
            ],
            expected_pseudo_time=datetime(2018, 3, 3, 12, 30, 0, tzinfo=timezone.utc),
        ),
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.DAEAD_ENCRYPTION_ALGORITHM.value,
                encryption_algorithm_parameters=[
                    {ENCRYPTION_PARAMETER_KEY_ID: "not-my-responsibility"}
                ],
            ),
            expected_algorithm=EncryptionAlgorithm.DAEAD_ENCRYPTION_ALGORITHM.value,
            expected_key=DAEAD_ENCRYPTION_KEY_REFERENCE,
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: "not-my-responsibility"},
            ],
            expected_pseudo_time=datetime(2025, 1, 1, 12, 30, 0, tzinfo=timezone.utc),
        ),
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm=EncryptionAlgorithm.DAEAD_ENCRYPTION_ALGORITHM.value,
                encryption_algorithm_parameters=[{"so-private": "key-hi-hi"}],
            ),
            expected_algorithm=EncryptionAlgorithm.DAEAD_ENCRYPTION_ALGORITHM.value,
            expected_key=DAEAD_ENCRYPTION_KEY_REFERENCE,
            expected_params=[
                {"so-private": "key-hi-hi"},
                {ENCRYPTION_PARAMETER_KEY_ID: DAEAD_ENCRYPTION_KEY_REFERENCE},
            ],
            expected_pseudo_time=datetime(2025, 1, 1, 12, 30, 0, tzinfo=timezone.utc),
        ),
    ],
    ids=[
        "Add new pseudonymization DAED",
        "Add pseudonymization DAED - encryption key refrence set",
        "No pseudonymization - empty pseudonymization",
        "Add pseudonymization DAED - pseudonymization time",
        "Add pseudonymization DAED - keyId in algorithm parameters",
        "Add pseudonymization DAED - additional algorithm parameters",
    ],
)
def test_add_default_pseudonymization_values_daed(case: PseudoCase, metadata: Datadoc):
    sykepenger = metadata.variables_lookup["sykepenger"]

    assert sykepenger.short_name is not None
    metadata.add_pseudonymization(sykepenger.short_name, case.new_pseudo)
    assert sykepenger.pseudonymization is not None

    assert sykepenger.pseudonymization.encryption_algorithm == case.expected_algorithm
    assert sykepenger.pseudonymization.encryption_key_reference == case.expected_key
    assert (
        sykepenger.pseudonymization.stable_identifier_version
        == case.expected_stable_version
    )
    algorithm_params = sykepenger.pseudonymization.encryption_algorithm_parameters
    if algorithm_params is None:
        assert case.expected_params is None
    else:
        _assert_dicts_in_list(
            case.expected_params,
            algorithm_params,
        )


@pytest.mark.parametrize(
    "case",
    [
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                encryption_algorithm="unknown",
            ),
            expected_algorithm="unknown",
        ),
        PseudoCase(
            new_pseudo=PseudonymizationOptional(
                pseudonymization_time=datetime(
                    2025, 10, 29, 0, 0, 0, tzinfo=timezone.utc
                ),
                encryption_algorithm="unknown",
                encryption_key_reference=DAEAD_ENCRYPTION_KEY_REFERENCE,
                stable_identifier_type=PAPIS_STABLE_IDENTIFIER_TYPE,
                stable_identifier_version="2-a-3",
                encryption_algorithm_parameters=[
                    {ENCRYPTION_PARAMETER_KEY_ID: PAPIS_ENCRYPTION_KEY_REFERENCE},
                    {"someKey": "specialValue"},
                ],
            ),
            expected_algorithm="unknown",
            expected_key=DAEAD_ENCRYPTION_KEY_REFERENCE,
            expected_stable_type=PAPIS_STABLE_IDENTIFIER_TYPE,
            expected_params=[
                {ENCRYPTION_PARAMETER_KEY_ID: PAPIS_ENCRYPTION_KEY_REFERENCE},
                {"someKey": "specialValue"},
            ],
            expected_pseudo_time=datetime(2025, 10, 29, 0, 0, 0, tzinfo=timezone.utc),
            expected_stable_version="2-a-3",
        ),
    ],
    ids=[
        "Add new unknown encryption algorithm",
        "Add new unknown encryption algorithm - all values",
    ],
)
def test_add_pseudonymization_unknown_algorithm(case: PseudoCase, metadata: Datadoc):
    sykepenger = metadata.variables_lookup["sykepenger"]
    assert sykepenger.short_name is not None

    metadata.add_pseudonymization(sykepenger.short_name, case.new_pseudo)
    assert sykepenger.pseudonymization is not None

    assert sykepenger.pseudonymization.encryption_algorithm == case.expected_algorithm
    assert sykepenger.pseudonymization.encryption_key_reference == case.expected_key
    assert (
        sykepenger.pseudonymization.stable_identifier_version
        == case.expected_stable_version
    )
    algorithm_params = sykepenger.pseudonymization.encryption_algorithm_parameters
    if algorithm_params is None:
        assert case.expected_params is None
    else:
        _assert_dicts_in_list(
            case.expected_params,
            algorithm_params,
        )


@pytest.mark.parametrize(
    "existing_metadata_path",
    [TEST_PSEUDO_DIRECTORY / "dataset_and_pseudo"],
)
def test_add_pseudo_variable(
    existing_metadata_file: Path,  # noqa: ARG001
    metadata: Datadoc,
):
    test_variable = "sykepenger"
    is_personal_data = metadata.variables_lookup[test_variable].is_personal_data
    metadata.add_pseudonymization(test_variable)
    assert metadata.variables_lookup[test_variable].pseudonymization is not None
    assert metadata.variables_lookup[test_variable].is_personal_data == is_personal_data


@pytest.mark.parametrize(
    "existing_metadata_path",
    [TEST_PSEUDO_DIRECTORY / "dataset_and_pseudo"],
)
def test_add_pseudo_variable_non_existent_variable_name(
    existing_metadata_file: Path,  # noqa: ARG001
    metadata: Datadoc,
):
    with pytest.raises(KeyError):
        metadata.add_pseudonymization("new_pseudo_variable")


@pytest.mark.parametrize(
    "existing_metadata_path",
    [TEST_PSEUDO_DIRECTORY / "dataset_and_pseudo"],
)
def test_existing_metadata_file_update_pseudonymization(
    existing_metadata_file: Path,  # noqa: ARG001
    metadata: Datadoc,
):
    metadata.add_pseudonymization("pers_id")
    variable = metadata.variables_lookup["pers_id"]

    assert variable.pseudonymization is not None
    assert variable.pseudonymization.encryption_algorithm is None
    variable.pseudonymization.encryption_algorithm = "new_encryption_algorithm"
    assert variable.pseudonymization.encryption_algorithm == "new_encryption_algorithm"


OPTIONAL_EXAMPLE = PseudonymizationOptional(
    encryption_algorithm=EncryptionAlgorithm.DAEAD_ENCRYPTION_ALGORITHM.value,
    encryption_key_reference="jippi-ya",
)
REQUIRED_EXAMPLE = PseudonymizationRequired(
    encryption_algorithm=EncryptionAlgorithm.DAEAD_ENCRYPTION_ALGORITHM.value,
    encryption_key_reference="jippi-ya",
)


@pytest.mark.parametrize(
    ("validate_required_fields_on_existing_metadata"),
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    ("input_model"),
    [
        OPTIONAL_EXAMPLE,
        REQUIRED_EXAMPLE,
    ],
)
def test_add_pseudonymization_model_types_no_existing_metadata(
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
    tmp_path: UPath,
    validate_required_fields_on_existing_metadata: bool,
    input_model: PseudonymizationType,
):
    shutil.copy(str(TEST_PARQUET_FILEPATH), str(tmp_path / TEST_PARQUET_FILE_NAME))
    metadata = Datadoc(
        str(tmp_path / TEST_PARQUET_FILE_NAME),
        statistic_subject_mapping=subject_mapping_fake_statistical_structure,
        validate_required_fields_on_existing_metadata=validate_required_fields_on_existing_metadata,
    )
    metadata.add_pseudonymization("pers_id", input_model)


@pytest.mark.parametrize(
    ("validate_required_fields_on_existing_metadata"),
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    ("input_model"),
    [
        OPTIONAL_EXAMPLE,
        REQUIRED_EXAMPLE,
    ],
)
def test_add_pseudonymization_model_types_existing_metadata(
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
    tmp_path: UPath,
    validate_required_fields_on_existing_metadata: bool,
    input_model: PseudonymizationType,
):
    shutil.copy(
        str(TEST_EXISTING_METADATA_NAMING_STANDARD_FILEPATH),
        str(tmp_path / TEST_EXISTING_METADATA_FILE_NAME),
    )
    shutil.copy(str(TEST_PARQUET_FILEPATH), str(tmp_path / TEST_PARQUET_FILE_NAME))
    metadata = Datadoc(
        str(tmp_path / TEST_PARQUET_FILE_NAME),
        statistic_subject_mapping=subject_mapping_fake_statistical_structure,
        validate_required_fields_on_existing_metadata=validate_required_fields_on_existing_metadata,
    )

    metadata.add_pseudonymization("fnr", input_model)


def test_add_pseudonymization_none_validate(
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
    tmp_path: UPath,
):
    shutil.copy(str(TEST_PARQUET_FILEPATH), str(tmp_path / TEST_PARQUET_FILE_NAME))
    metadata = Datadoc(
        str(tmp_path / TEST_PARQUET_FILE_NAME),
        statistic_subject_mapping=subject_mapping_fake_statistical_structure,
        validate_required_fields_on_existing_metadata=True,
    )
    with pytest.raises(ValueError, match="Can't add empty pseudonymization object"):
        metadata.add_pseudonymization("pers_id")


def test_add_pseudonymization_none(
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
    tmp_path: UPath,
):
    shutil.copy(str(TEST_PARQUET_FILEPATH), str(tmp_path / TEST_PARQUET_FILE_NAME))
    metadata = Datadoc(
        str(tmp_path / TEST_PARQUET_FILE_NAME),
        statistic_subject_mapping=subject_mapping_fake_statistical_structure,
        validate_required_fields_on_existing_metadata=False,
    )
    metadata.add_pseudonymization("pers_id")
    assert isinstance(
        metadata.variables_lookup["pers_id"].pseudonymization, PseudonymizationOptional
    )


@pytest.mark.parametrize(
    "existing_metadata_path",
    [TEST_PSEUDO_DIRECTORY / "dataset_and_pseudo"],
)
def test_update_pseudo(
    existing_metadata_file: Path,  # noqa: ARG001
    metadata: Datadoc,
):
    variable = metadata.variables_lookup["pers_id"]

    pseudo = PseudonymizationOptional(encryption_algorithm="new_encryption_algorithm")

    metadata.add_pseudonymization("pers_id", pseudo)

    assert variable.pseudonymization is not None
    assert variable.pseudonymization.encryption_algorithm == "new_encryption_algorithm"
    assert variable.pseudonymization.encryption_algorithm_parameters is None


@pytest.mark.parametrize(
    "existing_metadata_path",
    [TEST_PSEUDO_DIRECTORY / "dataset_and_pseudo"],
)
def test_remove_pseudo_variable(
    existing_metadata_file: Path,  # noqa: ARG001
    metadata: Datadoc,
):
    test_variable = "alm_inntekt"
    is_personal_data = metadata.variables_lookup[test_variable].is_personal_data
    metadata.remove_pseudonymization(test_variable)
    assert metadata.variables_lookup[test_variable].is_personal_data == is_personal_data
    assert metadata.variables_lookup[test_variable].pseudonymization is None


@pytest.mark.parametrize(
    "existing_metadata_path",
    [TEST_PSEUDO_DIRECTORY / "dataset_and_pseudo"],
)
def test_remove_pseudo_variable_non_existent_variable_name(
    existing_metadata_file: Path,  # noqa: ARG001
    metadata: Datadoc,
):
    with pytest.raises(KeyError):
        metadata.remove_pseudonymization("fnr")



"""Internal Variable Definitions Administration API

## Introduction  Variable Definitions are centralized definitions of concrete variables which are typically present in multiple datasets. Variable Definitions support standardization of data and metadata and facilitate sharing and joining of data by clarifying when variables have an identical definition.  ## Maintenance of Variable Definitions This API allows for creation, maintenance and access of Variable Definitions.  ### Ownership Creation and maintenance of variables may only be performed by Statistics Norway employees representing a specific Dapla team, who are defined as the owners of a given Variable Definition. The team an owner represents must be specified when making a request through the `active_group` query parameter. All maintenance is to be performed by the owners, with no intervention from administrators.  ### Status All Variable Definitions have an associated status. The possible values for status are `DRAFT`, `PUBLISHED_INTERNAL` and `PUBLISHED_EXTERNAL`.  #### Draft When a Variable Definition is created it is assigned the status `DRAFT`. Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Mutable (it may be changed directly without need for versioning). - Not suitable to refer to from other systems.  This status may be changed to `PUBLISHED_INTERNAL` or `PUBLISHED_EXTERNAL` with a direct update.  #### Published Internal Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Immutable (all changes are versioned). - Suitable to refer to in internal systems for statistics production. - Not suitable to refer to for external use (for example in Statistikkbanken).  This status may be changed to `PUBLISHED_EXTERNAL` by creating a Patch version.  #### Published External Under this status the Variable Definition is:  - Visible to the general public. - Immutable (all changes are versioned). - Suitable to refer to from any system.  This status may not be changed as it would break immutability. If a Variable Definition is no longer relevant then its period of validity should be ended by specifying a `valid_until` date in a Patch version.  ### Immutability Variable Definitions are immutable. This means that any changes must be performed in a strict versioning system. Consumers can avoid being exposed to breaking changes by specifying a `date_of_validity` when they request a Variable Definition.  #### Patches Patches are for changes which do not affect the fundamental meaning of the Variable Definition.  #### Validity Periods Validity Periods are versions with a period defined by a `valid_from` date and optionally a `valid_until` date. If the fundamental meaning of a Variable Definition is to be changed, it should be done by creating a new Validity Period.

The version of the OpenAPI document: 0.1
Contact: metadata@ssb.no
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""

from datetime import date
from typing import Annotated
from typing import Any

from pydantic import Field
from pydantic import StrictFloat
from pydantic import StrictInt
from pydantic import StrictStr
from pydantic import validate_call

from ..api_client import ApiClient
from ..api_client import RequestSerialized
from ..api_response import ApiResponse
from ..models.complete_response import CompleteResponse
from ..models.patch import Patch
from ..rest import RESTResponseType


class PatchesApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client

    @validate_call
    def create_patch(
        self,
        variable_definition_id: Annotated[
            StrictStr,
            Field(description="Unique identifier for the variable definition."),
        ],
        valid_from: Annotated[
            date | None,
            Field(
                description="Valid from date for the specific validity period to be patched."
            ),
        ] = None,
        patch: Patch | None = None,
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> CompleteResponse:
        """Create a new patch for a variable definition.

        Create a new patch for a variable definition.

        :param variable_definition_id: Unique identifier for the variable definition. (required)
        :type variable_definition_id: str
        :param valid_from: Valid from date for the specific validity period to be patched.
        :type valid_from: date
        :param patch:
        :type patch: Patch
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._create_patch_serialize(
            variable_definition_id=variable_definition_id,
            valid_from=valid_from,
            patch=patch,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "201": "CompleteResponse",
            "404": "Problem",
            "400": "Problem",
            "405": "Problem",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def create_patch_with_http_info(
        self,
        variable_definition_id: Annotated[
            StrictStr,
            Field(description="Unique identifier for the variable definition."),
        ],
        valid_from: Annotated[
            date | None,
            Field(
                description="Valid from date for the specific validity period to be patched."
            ),
        ] = None,
        patch: Patch | None = None,
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[CompleteResponse]:
        """Create a new patch for a variable definition.

        Create a new patch for a variable definition.

        :param variable_definition_id: Unique identifier for the variable definition. (required)
        :type variable_definition_id: str
        :param valid_from: Valid from date for the specific validity period to be patched.
        :type valid_from: date
        :param patch:
        :type patch: Patch
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._create_patch_serialize(
            variable_definition_id=variable_definition_id,
            valid_from=valid_from,
            patch=patch,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "201": "CompleteResponse",
            "404": "Problem",
            "400": "Problem",
            "405": "Problem",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def create_patch_without_preload_content(
        self,
        variable_definition_id: Annotated[
            StrictStr,
            Field(description="Unique identifier for the variable definition."),
        ],
        valid_from: Annotated[
            date | None,
            Field(
                description="Valid from date for the specific validity period to be patched."
            ),
        ] = None,
        patch: Patch | None = None,
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Create a new patch for a variable definition.

        Create a new patch for a variable definition.

        :param variable_definition_id: Unique identifier for the variable definition. (required)
        :type variable_definition_id: str
        :param valid_from: Valid from date for the specific validity period to be patched.
        :type valid_from: date
        :param patch:
        :type patch: Patch
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._create_patch_serialize(
            variable_definition_id=variable_definition_id,
            valid_from=valid_from,
            patch=patch,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "201": "CompleteResponse",
            "404": "Problem",
            "400": "Problem",
            "405": "Problem",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        return response_data.response

    def _create_patch_serialize(
        self,
        variable_definition_id,
        valid_from,
        patch,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:
        _host = None

        _collection_formats: dict[str, str] = {}

        _path_params: dict[str, str] = {}
        _query_params: list[tuple[str, str]] = []
        _header_params: dict[str, str | None] = _headers or {}
        _form_params: list[tuple[str, str]] = []
        _files: dict[
            str, str | bytes | list[str] | list[bytes] | list[tuple[str, bytes]]
        ] = {}
        _body_params: bytes | None = None

        # process the path parameters
        if variable_definition_id is not None:
            _path_params["variable-definition-id"] = variable_definition_id
        # process the query parameters
        if valid_from is not None:
            if isinstance(valid_from, date):
                _query_params.append(
                    (
                        "valid_from",
                        valid_from.strftime(self.api_client.configuration.date_format),
                    )
                )
            else:
                _query_params.append(("valid_from", valid_from))

        # process the header parameters
        # process the form parameters
        # process the body parameter
        if patch is not None:
            _body_params = patch

        # set the HTTP header `Accept`
        if "Accept" not in _header_params:
            _header_params["Accept"] = self.api_client.select_header_accept(
                ["application/json", "application/problem+json"]
            )

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params["Content-Type"] = _content_type
        else:
            _default_content_type = self.api_client.select_header_content_type(
                ["application/json"]
            )
            if _default_content_type is not None:
                _header_params["Content-Type"] = _default_content_type

        # authentication setting
        _auth_settings: list[str] = ["labid_token"]

        return self.api_client.param_serialize(
            method="POST",
            resource_path="/variable-definitions/{variable-definition-id}/patches",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth,
        )

    @validate_call
    def get_patch(
        self,
        variable_definition_id: Annotated[
            StrictStr,
            Field(description="Unique identifier for the variable definition."),
        ],
        patch_id: Annotated[
            StrictInt, Field(description="ID of the patch to retrieve")
        ],
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> CompleteResponse:
        """Get one concrete patch for the given variable definition.

        Get one concrete patch for the given variable definition. The full object is returned for comparison purposes.

        :param variable_definition_id: Unique identifier for the variable definition. (required)
        :type variable_definition_id: str
        :param patch_id: ID of the patch to retrieve (required)
        :type patch_id: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._get_patch_serialize(
            variable_definition_id=variable_definition_id,
            patch_id=patch_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "200": "CompleteResponse",
            "404": "Problem",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def get_patch_with_http_info(
        self,
        variable_definition_id: Annotated[
            StrictStr,
            Field(description="Unique identifier for the variable definition."),
        ],
        patch_id: Annotated[
            StrictInt, Field(description="ID of the patch to retrieve")
        ],
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[CompleteResponse]:
        """Get one concrete patch for the given variable definition.

        Get one concrete patch for the given variable definition. The full object is returned for comparison purposes.

        :param variable_definition_id: Unique identifier for the variable definition. (required)
        :type variable_definition_id: str
        :param patch_id: ID of the patch to retrieve (required)
        :type patch_id: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._get_patch_serialize(
            variable_definition_id=variable_definition_id,
            patch_id=patch_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "200": "CompleteResponse",
            "404": "Problem",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def get_patch_without_preload_content(
        self,
        variable_definition_id: Annotated[
            StrictStr,
            Field(description="Unique identifier for the variable definition."),
        ],
        patch_id: Annotated[
            StrictInt, Field(description="ID of the patch to retrieve")
        ],
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Get one concrete patch for the given variable definition.

        Get one concrete patch for the given variable definition. The full object is returned for comparison purposes.

        :param variable_definition_id: Unique identifier for the variable definition. (required)
        :type variable_definition_id: str
        :param patch_id: ID of the patch to retrieve (required)
        :type patch_id: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._get_patch_serialize(
            variable_definition_id=variable_definition_id,
            patch_id=patch_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "200": "CompleteResponse",
            "404": "Problem",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        return response_data.response

    def _get_patch_serialize(
        self,
        variable_definition_id,
        patch_id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:
        _host = None

        _collection_formats: dict[str, str] = {}

        _path_params: dict[str, str] = {}
        _query_params: list[tuple[str, str]] = []
        _header_params: dict[str, str | None] = _headers or {}
        _form_params: list[tuple[str, str]] = []
        _files: dict[
            str, str | bytes | list[str] | list[bytes] | list[tuple[str, bytes]]
        ] = {}
        _body_params: bytes | None = None

        # process the path parameters
        if variable_definition_id is not None:
            _path_params["variable-definition-id"] = variable_definition_id
        if patch_id is not None:
            _path_params["patch-id"] = patch_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if "Accept" not in _header_params:
            _header_params["Accept"] = self.api_client.select_header_accept(
                ["application/json", "application/problem+json"]
            )

        # authentication setting
        _auth_settings: list[str] = ["labid_token"]

        return self.api_client.param_serialize(
            method="GET",
            resource_path="/variable-definitions/{variable-definition-id}/patches/{patch-id}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth,
        )

    @validate_call
    def list_patches(
        self,
        variable_definition_id: Annotated[
            StrictStr,
            Field(description="Unique identifier for the variable definition."),
        ],
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> list[CompleteResponse]:
        """List all patches for the given variable definition.

        List all patches for the given variable definition. The full object is returned for comparison purposes.

        :param variable_definition_id: Unique identifier for the variable definition. (required)
        :type variable_definition_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._list_patches_serialize(
            variable_definition_id=variable_definition_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "200": "List[CompleteResponse]",
            "404": "Problem",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def list_patches_with_http_info(
        self,
        variable_definition_id: Annotated[
            StrictStr,
            Field(description="Unique identifier for the variable definition."),
        ],
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[list[CompleteResponse]]:
        """List all patches for the given variable definition.

        List all patches for the given variable definition. The full object is returned for comparison purposes.

        :param variable_definition_id: Unique identifier for the variable definition. (required)
        :type variable_definition_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._list_patches_serialize(
            variable_definition_id=variable_definition_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "200": "List[CompleteResponse]",
            "404": "Problem",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def list_patches_without_preload_content(
        self,
        variable_definition_id: Annotated[
            StrictStr,
            Field(description="Unique identifier for the variable definition."),
        ],
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """List all patches for the given variable definition.

        List all patches for the given variable definition. The full object is returned for comparison purposes.

        :param variable_definition_id: Unique identifier for the variable definition. (required)
        :type variable_definition_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._list_patches_serialize(
            variable_definition_id=variable_definition_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "200": "List[CompleteResponse]",
            "404": "Problem",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        return response_data.response

    def _list_patches_serialize(
        self,
        variable_definition_id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:
        _host = None

        _collection_formats: dict[str, str] = {}

        _path_params: dict[str, str] = {}
        _query_params: list[tuple[str, str]] = []
        _header_params: dict[str, str | None] = _headers or {}
        _form_params: list[tuple[str, str]] = []
        _files: dict[
            str, str | bytes | list[str] | list[bytes] | list[tuple[str, bytes]]
        ] = {}
        _body_params: bytes | None = None

        # process the path parameters
        if variable_definition_id is not None:
            _path_params["variable-definition-id"] = variable_definition_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if "Accept" not in _header_params:
            _header_params["Accept"] = self.api_client.select_header_accept(
                ["application/json", "application/problem+json"]
            )

        # authentication setting
        _auth_settings: list[str] = ["labid_token"]

        return self.api_client.param_serialize(
            method="GET",
            resource_path="/variable-definitions/{variable-definition-id}/patches",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth,
        )



# flake8: noqa

# import apis into api package
from ..api.data_migration_api import DataMigrationApi
from ..api.draft_variable_definitions_api import DraftVariableDefinitionsApi
from ..api.patches_api import PatchesApi
from ..api.validity_periods_api import ValidityPeriodsApi
from ..api.variable_definitions_api import VariableDefinitionsApi



from __future__ import annotations

import datetime  # import is needed in xdoctest
import logging
import uuid
from typing import Any
from typing import TypeAlias

import datadoc_model.all_optional.model as all_optional_model
import datadoc_model.required.model as required_model
from datadoc_model.all_optional.model import Assessment
from datadoc_model.all_optional.model import DataSetState
from datadoc_model.all_optional.model import VariableRole

from dapla_metadata.dapla import user_info
from dapla_metadata.datasets.utility.constants import DAEAD_ENCRYPTION_KEY_REFERENCE
from dapla_metadata.datasets.utility.constants import ENCRYPTION_PARAMETER_KEY_ID
from dapla_metadata.datasets.utility.constants import ENCRYPTION_PARAMETER_SNAPSHOT_DATE
from dapla_metadata.datasets.utility.constants import ENCRYPTION_PARAMETER_STRATEGY
from dapla_metadata.datasets.utility.constants import ENCRYPTION_PARAMETER_STRATEGY_SKIP
from dapla_metadata.datasets.utility.constants import NUM_OBLIGATORY_VARIABLES_FIELDS
from dapla_metadata.datasets.utility.constants import (
    OBLIGATORY_DATASET_METADATA_IDENTIFIERS,
)
from dapla_metadata.datasets.utility.constants import (
    OBLIGATORY_DATASET_METADATA_IDENTIFIERS_MULTILANGUAGE,
)
from dapla_metadata.datasets.utility.constants import (
    OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS,
)
from dapla_metadata.datasets.utility.constants import (
    OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS_MULTILANGUAGE,
)
from dapla_metadata.datasets.utility.constants import (
    OBLIGATORY_VARIABLES_PSEUDONYMIZATION_IDENTIFIERS,
)
from dapla_metadata.datasets.utility.constants import PAPIS_ENCRYPTION_KEY_REFERENCE
from dapla_metadata.datasets.utility.constants import PAPIS_STABLE_IDENTIFIER_TYPE
from dapla_metadata.datasets.utility.enums import EncryptionAlgorithm

logger = logging.getLogger(__name__)

DatadocMetadataType: TypeAlias = (
    all_optional_model.DatadocMetadata | required_model.DatadocMetadata
)
DatasetType: TypeAlias = all_optional_model.Dataset | required_model.Dataset
VariableType: TypeAlias = all_optional_model.Variable | required_model.Variable
PseudonymizationType: TypeAlias = (
    all_optional_model.Pseudonymization | required_model.Pseudonymization
)
VariableListType: TypeAlias = (
    list[all_optional_model.Variable] | list[required_model.Variable]
)
OptionalDatadocMetadataType: TypeAlias = DatadocMetadataType | None


def get_current_date() -> str:
    """Return a current date as str."""
    return datetime.datetime.now(tz=datetime.timezone.utc).date().isoformat()


def get_timestamp_now() -> datetime.datetime:
    """Return a timestamp for the current moment."""
    return datetime.datetime.now(tz=datetime.timezone.utc)


def calculate_percentage(completed: int, total: int) -> int:
    """Calculate percentage as a rounded integer.

    Args:
        completed: The number of completed items.
        total: The total number of items.

    Returns:
        The rounded percentage of completed items out of the total.
    """
    return round((completed / total) * 100)


def derive_assessment_from_state(state: DataSetState) -> Assessment:
    """Derive assessment from dataset state.

    Args:
        state: The state of the dataset.

    Returns:
        The derived assessment of the dataset.
    """
    match state:
        case (
            DataSetState.INPUT_DATA
            | DataSetState.PROCESSED_DATA
            | DataSetState.STATISTICS
        ):
            return Assessment.PROTECTED
        case DataSetState.OUTPUT_DATA:
            return Assessment.OPEN
        case DataSetState.SOURCE_DATA:
            return Assessment.SENSITIVE


def set_default_values_variables(variables: VariableListType) -> None:
    """Set default values on variables.

    Args:
        variables: A list of variable objects to set default values on.

    Example:
        >>> variables = [all_optional_model.Variable(short_name="pers",id=None, is_personal_data = None), all_optional_model.Variable(short_name="fnr",id='9662875c-c245-41de-b667-12ad2091a1ee', is_personal_data=True)]
        >>> set_default_values_variables(variables)
        >>> isinstance(variables[0].id, uuid.UUID)
        True

        >>> variables[1].is_personal_data == True
        True

        >>> variables[0].is_personal_data == False
        True
    """
    for v in variables:
        if v.id is None:
            v.id = uuid.uuid4()
        if v.is_personal_data is None:
            v.is_personal_data = False
        if v.variable_role is None:
            v.variable_role = VariableRole.MEASURE


def set_default_values_dataset(
    dataset: DatasetType,
) -> None:
    """Set default values on dataset.

    Args:
        dataset: The dataset object to set default values on.

    Example:
        >>> dataset = all_optional_model.Dataset(id=None)
        >>> set_default_values_dataset(dataset)
        >>> dataset.id is not None
        True
    """
    if not dataset.id:
        dataset.id = uuid.uuid4()


def set_dataset_owner(
    dataset: DatasetType,
) -> None:
    """Sets the owner of the dataset from the DAPLA_GROUP_CONTEXT enviornment variable.

    Args:
        dataset: The dataset object to set default values on.
    """
    try:
        dataset.owner = user_info.get_user_info_for_current_platform().current_team
    except OSError:
        logger.exception("Failed to find environment variable DAPLA_GROUP_CONTEXT")


def set_variables_inherit_from_dataset(
    dataset: DatasetType,
    variables: list,
) -> None:
    """Set specific dataset values on a list of variable objects.

    This function populates 'data source', 'temporality type', 'contains data from',
    and 'contains data until' fields in each variable if they are not set (None).
    The values are inherited from the corresponding fields in the dataset.

    Args:
        dataset: The dataset object from which to inherit values.
        variables: A list of variable objects to update with dataset values.

    Example:
        >>> dataset = all_optional_model.Dataset(short_name='person_data_v1', id='9662875c-c245-41de-b667-12ad2091a1ee', contains_data_from="2010-09-05", contains_data_until="2022-09-05")
        >>> variables = [all_optional_model.Variable(short_name="pers", data_source=None, temporality_type=None, contains_data_from=None, contains_data_until=None)]
        >>> set_variables_inherit_from_dataset(dataset, variables)

        >>> variables[0].contains_data_from == dataset.contains_data_from
        True

        >>> variables[0].contains_data_until == dataset.contains_data_until
        True
    """
    for v in variables:
        v.contains_data_from = v.contains_data_from or dataset.contains_data_from
        v.contains_data_until = v.contains_data_until or dataset.contains_data_until


def incorrect_date_order(
    date_from: datetime.date | None,
    date_until: datetime.date | None,
) -> bool:
    """Evaluate the chronological order of two dates.

    This function checks if 'date until' is earlier than 'date from'. If so, it
    indicates an incorrect date order.

    Args:
        date_from: The start date of the time period.
        date_until: The end date of the time period.

    Returns:
        True if 'date_until' is earlier than 'date_from' or if only 'date_from' is None, False otherwise.

    Example:
        >>> incorrect_date_order(datetime.date(1980, 1, 1), datetime.date(1967, 1, 1))
        True

        >>> incorrect_date_order(datetime.date(1967, 1, 1), datetime.date(1980, 1, 1))
        False

        >>> incorrect_date_order(None, datetime.date(2024,7,1))
        True
    """
    if date_from is None and date_until is not None:
        return True
    return date_from is not None and date_until is not None and date_until < date_from


def _is_missing_multilanguage_value(
    field_name: str,
    field_value,  # noqa: ANN001 Skip type hint to enable dynamically handling value for LanguageStringType not indexable
    obligatory_list: list,
) -> bool:
    """Check obligatory fields with multilanguage value.

    This function checks whether a given field, which is supposed to have
    multilanguage values, is missing values in all specified languages.

    Args:
        field_name: The name of the field to check.
        field_value: The value of the field. Expected to be of type LanguageStringType.
        obligatory_list: A list of obligatory field names that should have multilanguage values.

    Returns:
        True if no value in any of languages for one field, False otherwise.
    """
    return bool(
        field_name in obligatory_list
        and field_value
        and (
            len(field_value[0]) > 0
            and not field_value[0]["languageText"]
            and (len(field_value) <= 1 or not field_value[1]["languageText"])
            and (len(field_value) <= 2 or not field_value[2]["languageText"])
        ),
    )


def _is_missing_metadata(
    field_name: str,
    field_value,  # noqa: ANN001 Skip type hint because method '_is_missing_multilanguage_value'
    obligatory_list: list,
    obligatory_multi_language_list: list,
) -> bool:
    """Check if an obligatory field is missing its value.

    This function checks whether a given field, which may be a simple string or a
    multilanguage value, is missing its value. It considers two lists: one for
    obligatory fields and another for obligatory multilanguage fields.

    Args:
        field_name: The name of the field to check.
        field_value: The value of the field. Can be of type str, or LanguageStringType for
            multilanguage fields.
        obligatory_list: List of obligatory fields.
        obligatory_multi_language_list: List of obligatory fields with multilanguage
            values.

    Returns:
        True if the field doesn't have a value, False otherwise.
    """
    return bool(
        (field_name in obligatory_list and field_value is None)
        or _is_missing_multilanguage_value(
            field_name,
            field_value,
            obligatory_multi_language_list,
        ),
    )


def num_obligatory_dataset_fields_completed(
    dataset: DatasetType,
) -> int:
    """Count the number of completed obligatory dataset fields.

    This function returns the total count of obligatory fields in the dataset that
    have values (are not None).

    Args:
        dataset: The dataset object for which to count the fields.

    Returns:
        The number of obligatory dataset fields that have been completed (not None).
    """
    return len(OBLIGATORY_DATASET_METADATA_IDENTIFIERS) - len(
        get_missing_obligatory_dataset_fields(dataset),
    )


def num_obligatory_variables_fields_completed(variables: list) -> int:
    """Count the number of obligatory fields completed for all variables.

    This function calculates the total number of obligatory fields that have
    values (are not None) for one variable in the list.

    Args:
        variables: A list with variable objects.

    Returns:
        The total number of obligatory variable fields that have been completed
        (not None) for all variables.
    """
    num_completed = 0
    for v in variables:
        num_completed += num_obligatory_variable_fields_completed(v)
    return num_completed


def num_obligatory_variable_fields_completed(
    variable: all_optional_model.Variable,
) -> int:
    """Count the number of obligatory fields completed for one variable.

    This function calculates the total number of obligatory fields that have
    values (are not None) for one variable in the list.

    Args:
        variable: The variable to count obligatory fields for.

    Returns:
        The total number of obligatory variable fields that have been completed
        (not None) for one variable.
    """
    missing_metadata = [
        key
        for key, value in variable.model_dump().items()
        if _is_missing_metadata(
            key,
            value,
            OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS,
            OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS_MULTILANGUAGE,
        )
    ]
    return NUM_OBLIGATORY_VARIABLES_FIELDS - len(missing_metadata)


def num_obligatory_pseudo_fields_missing(
    variables: list[all_optional_model.Variable],
) -> int:
    """Counts the number of obligatory pseudonymization fields are missing.

    Args:
        variables: The variables to count obligatory fields for.

    Returns:
        The number of obligatory pseudonymization fields that are missing.
    """
    return sum(
        getattr(v.pseudonymization, field, None) is None
        for v in variables
        if v.pseudonymization is not None
        for field in OBLIGATORY_VARIABLES_PSEUDONYMIZATION_IDENTIFIERS
    )


def get_missing_obligatory_dataset_fields(
    dataset: DatasetType,
) -> list:
    """Identify all obligatory dataset fields that are missing values.

    This function checks for obligatory fields that are either directly missing
    (i.e., set to `None`) or have multilanguage values with empty content.

    Args:
        dataset: The dataset object to examine. This object must support the
            `model_dump()` method which returns a dictionary of field names and
            values.

    Returns:
        A list of field names (as strings) that are missing values. This includes:

        - Fields that are directly `None` and are listed as obligatory metadata.
        - Multilanguage fields (listed as obligatory metadata`) where
            the value exists but the primary language text is empty.

    """
    return [
        key
        for key, value in dataset.model_dump().items()
        if _is_missing_metadata(
            key,
            value,
            OBLIGATORY_DATASET_METADATA_IDENTIFIERS,
            OBLIGATORY_DATASET_METADATA_IDENTIFIERS_MULTILANGUAGE,
        )
    ]


def get_missing_obligatory_variables_fields(variables: list) -> list[dict]:
    """Identify obligatory variable fields that are missing values for each variable.

    This function checks for obligatory fields that are either directly missing
    (i.e., set to `None`) or have multilanguage values with empty content.

    Args:
        variables: A list of variable objects to check for missing obligatory fields.

    Returns:
        A list of dictionaries with variable short names as keys and list of missing
        obligatory variable fields as values. This includes:
        - Fields that are directly `None` and are llisted as obligatory metadata.
        - Multilanguage fields (listed as obligatory metadata) where the value
            exists but the primary language text is empty.

    """
    missing_variables_fields = [
        {
            variable.short_name: [
                key
                for key, value in variable.model_dump().items()
                if _is_missing_metadata(
                    key,
                    value,
                    OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS,
                    OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS_MULTILANGUAGE,
                )
            ],
        }
        for variable in variables
    ]
    # Filtering out variable keys with empty values list
    return [item for item in missing_variables_fields if next(iter(item.values()))]


def get_missing_obligatory_variables_pseudo_fields(
    variables: list[all_optional_model.Variable],
) -> list[dict]:
    """Identify obligatory variable pseudonymization fields that are missing values for each variable.

    This function checks for obligatory fields that are directly missing
    (i.e., set to `None`).

    Args:
        variables: A list of variable objects to check for missing obligatory pseudonymization fields.

    Returns:
        A list of dictionaries with variable short names as keys and list of missing
        obligatory variable pseudonymization fields as values. This includes:
        - Fields that are directly `None` and are listed as obligatory metadata.

    """
    return [
        {
            v.short_name: [
                key
                for key, value in v.pseudonymization.model_dump().items()
                if _is_missing_metadata(
                    key,
                    value,
                    OBLIGATORY_VARIABLES_PSEUDONYMIZATION_IDENTIFIERS,
                    [],
                )
            ]
        }
        for v in variables
        if v.pseudonymization is not None
    ]


def running_in_notebook() -> bool:
    """Return True if running in Jupyter Notebook."""
    try:
        return bool(get_ipython().__class__.__name__ == "ZMQInteractiveShell")  # type: ignore [name-defined]
    except NameError:
        # The get_ipython method is globally available in ipython interpreters
        # as used in Jupyter. However it is not available in other python
        # interpreters and will throw a NameError. Therefore we're not running
        # in Jupyter.
        return False


def _ensure_encryption_parameters(
    existing: list[dict[str, Any]] | None,
    required: dict[str, Any],
) -> list[dict[str, Any]]:
    """Ensure required key/value pairs exist in parameters list."""
    result = list(existing or [])

    # Ensure each required key is present in at least one dict
    for key, value in required.items():
        if not any(key in d for d in result):
            result.append({key: value})

    return result


def set_default_values_pseudonymization(
    variable: VariableType,
    pseudonymization: PseudonymizationType,
) -> None:
    """Populate pseudonymization fields with defaults based on the encryption algorithm.

    Updates the encryption key reference and encryption parameters if they are not set,
    handling both PAPIS and DAED algorithms. Leaves unknown algorithms unchanged.
    """
    # Use the model corresponding to that already in use internally
    if isinstance(variable, required_model.Variable) and isinstance(
        pseudonymization, all_optional_model.Pseudonymization
    ):
        pseudonymization = required_model.Pseudonymization.model_validate(
            pseudonymization.model_dump()
        )
    elif isinstance(variable, all_optional_model.Variable) and isinstance(
        pseudonymization, required_model.Pseudonymization
    ):
        pseudonymization = all_optional_model.Pseudonymization.model_validate(
            pseudonymization.model_dump()
        )
    if variable.pseudonymization is None:
        variable.pseudonymization = pseudonymization
    match pseudonymization.encryption_algorithm:
        case EncryptionAlgorithm.PAPIS_ENCRYPTION_ALGORITHM.value:
            if not pseudonymization.encryption_key_reference:
                pseudonymization.encryption_key_reference = (
                    PAPIS_ENCRYPTION_KEY_REFERENCE
                )
            base_params = {
                ENCRYPTION_PARAMETER_KEY_ID: PAPIS_ENCRYPTION_KEY_REFERENCE,
                ENCRYPTION_PARAMETER_STRATEGY: ENCRYPTION_PARAMETER_STRATEGY_SKIP,
            }
            if pseudonymization.stable_identifier_type == PAPIS_STABLE_IDENTIFIER_TYPE:
                base_params[ENCRYPTION_PARAMETER_SNAPSHOT_DATE] = get_current_date()
            pseudonymization.encryption_algorithm_parameters = (
                _ensure_encryption_parameters(
                    pseudonymization.encryption_algorithm_parameters,
                    base_params,
                )
            )
        case EncryptionAlgorithm.DAEAD_ENCRYPTION_ALGORITHM.value:
            if not pseudonymization.encryption_key_reference:
                pseudonymization.encryption_key_reference = (
                    DAEAD_ENCRYPTION_KEY_REFERENCE
                )
            pseudonymization.encryption_algorithm_parameters = (
                _ensure_encryption_parameters(
                    pseudonymization.encryption_algorithm_parameters,
                    {
                        ENCRYPTION_PARAMETER_KEY_ID: DAEAD_ENCRYPTION_KEY_REFERENCE,
                    },
                )
            )
        case _:
            pass



"""Internal Variable Definitions Administration API

## Introduction  Variable Definitions are centralized definitions of concrete variables which are typically present in multiple datasets. Variable Definitions support standardization of data and metadata and facilitate sharing and joining of data by clarifying when variables have an identical definition.  ## Maintenance of Variable Definitions This API allows for creation, maintenance and access of Variable Definitions.  ### Ownership Creation and maintenance of variables may only be performed by Statistics Norway employees representing a specific Dapla team, who are defined as the owners of a given Variable Definition. The team an owner represents must be specified when making a request through the `active_group` query parameter. All maintenance is to be performed by the owners, with no intervention from administrators.  ### Status All Variable Definitions have an associated status. The possible values for status are `DRAFT`, `PUBLISHED_INTERNAL` and `PUBLISHED_EXTERNAL`.  #### Draft When a Variable Definition is created it is assigned the status `DRAFT`. Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Mutable (it may be changed directly without need for versioning). - Not suitable to refer to from other systems.  This status may be changed to `PUBLISHED_INTERNAL` or `PUBLISHED_EXTERNAL` with a direct update.  #### Published Internal Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Immutable (all changes are versioned). - Suitable to refer to in internal systems for statistics production. - Not suitable to refer to for external use (for example in Statistikkbanken).  This status may be changed to `PUBLISHED_EXTERNAL` by creating a Patch version.  #### Published External Under this status the Variable Definition is:  - Visible to the general public. - Immutable (all changes are versioned). - Suitable to refer to from any system.  This status may not be changed as it would break immutability. If a Variable Definition is no longer relevant then its period of validity should be ended by specifying a `valid_until` date in a Patch version.  ### Immutability Variable Definitions are immutable. This means that any changes must be performed in a strict versioning system. Consumers can avoid being exposed to breaking changes by specifying a `date_of_validity` when they request a Variable Definition.  #### Patches Patches are for changes which do not affect the fundamental meaning of the Variable Definition.  #### Validity Periods Validity Periods are versions with a period defined by a `valid_from` date and optionally a `valid_until` date. If the fundamental meaning of a Variable Definition is to be changed, it should be done by creating a new Validity Period.

The version of the OpenAPI document: 0.1
Contact: metadata@ssb.no
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""

from typing import Annotated
from typing import Any

from pydantic import Field
from pydantic import StrictFloat
from pydantic import StrictInt
from pydantic import StrictStr
from pydantic import validate_call

from ..api_client import ApiClient
from ..api_client import RequestSerialized
from ..api_response import ApiResponse
from ..models.complete_response import CompleteResponse
from ..models.get_vardok_vardef_mapping_by_id200_response import (
    GetVardokVardefMappingById200Response,
)
from ..models.vardok_vardef_id_pair_response import VardokVardefIdPairResponse
from ..rest import RESTResponseType


class DataMigrationApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client

    @validate_call
    def create_variable_definition_from_var_dok(
        self,
        vardok_id: Annotated[
            StrictStr, Field(description="The ID of the definition in Vardok.")
        ],
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> CompleteResponse:
        """Create a variable definition from a VarDok variable definition.

        Create a variable definition from a VarDok variable definition.

        :param vardok_id: The ID of the definition in Vardok. (required)
        :type vardok_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._create_variable_definition_from_var_dok_serialize(
            vardok_id=vardok_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "201": "CompleteResponse",
            "400": "Problem",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def create_variable_definition_from_var_dok_with_http_info(
        self,
        vardok_id: Annotated[
            StrictStr, Field(description="The ID of the definition in Vardok.")
        ],
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[CompleteResponse]:
        """Create a variable definition from a VarDok variable definition.

        Create a variable definition from a VarDok variable definition.

        :param vardok_id: The ID of the definition in Vardok. (required)
        :type vardok_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._create_variable_definition_from_var_dok_serialize(
            vardok_id=vardok_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "201": "CompleteResponse",
            "400": "Problem",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def create_variable_definition_from_var_dok_without_preload_content(
        self,
        vardok_id: Annotated[
            StrictStr, Field(description="The ID of the definition in Vardok.")
        ],
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Create a variable definition from a VarDok variable definition.

        Create a variable definition from a VarDok variable definition.

        :param vardok_id: The ID of the definition in Vardok. (required)
        :type vardok_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._create_variable_definition_from_var_dok_serialize(
            vardok_id=vardok_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "201": "CompleteResponse",
            "400": "Problem",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        return response_data.response

    def _create_variable_definition_from_var_dok_serialize(
        self,
        vardok_id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:
        _host = None

        _collection_formats: dict[str, str] = {}

        _path_params: dict[str, str] = {}
        _query_params: list[tuple[str, str]] = []
        _header_params: dict[str, str | None] = _headers or {}
        _form_params: list[tuple[str, str]] = []
        _files: dict[
            str, str | bytes | list[str] | list[bytes] | list[tuple[str, bytes]]
        ] = {}
        _body_params: bytes | None = None

        # process the path parameters
        if vardok_id is not None:
            _path_params["vardok-id"] = vardok_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if "Accept" not in _header_params:
            _header_params["Accept"] = self.api_client.select_header_accept(
                ["application/json", "application/problem+json"]
            )

        # authentication setting
        _auth_settings: list[str] = ["labid_token"]

        return self.api_client.param_serialize(
            method="POST",
            resource_path="/vardok-migration/{vardok-id}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth,
        )

    @validate_call
    def get_vardok_vardef_mapping(
        self,
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> list[VardokVardefIdPairResponse]:
        """Get a list of all vardok and vardef id mappings

        Get a list of all vardok and vardef id mappings

        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._get_vardok_vardef_mapping_serialize(
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "200": "List[VardokVardefIdPairResponse]",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def get_vardok_vardef_mapping_with_http_info(
        self,
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[list[VardokVardefIdPairResponse]]:
        """Get a list of all vardok and vardef id mappings

        Get a list of all vardok and vardef id mappings

        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._get_vardok_vardef_mapping_serialize(
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "200": "List[VardokVardefIdPairResponse]",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def get_vardok_vardef_mapping_without_preload_content(
        self,
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Get a list of all vardok and vardef id mappings

        Get a list of all vardok and vardef id mappings

        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._get_vardok_vardef_mapping_serialize(
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "200": "List[VardokVardefIdPairResponse]",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        return response_data.response

    def _get_vardok_vardef_mapping_serialize(
        self,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:
        _host = None

        _collection_formats: dict[str, str] = {}

        _path_params: dict[str, str] = {}
        _query_params: list[tuple[str, str]] = []
        _header_params: dict[str, str | None] = _headers or {}
        _form_params: list[tuple[str, str]] = []
        _files: dict[
            str, str | bytes | list[str] | list[bytes] | list[tuple[str, bytes]]
        ] = {}
        _body_params: bytes | None = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if "Accept" not in _header_params:
            _header_params["Accept"] = self.api_client.select_header_accept(
                ["application/json"]
            )

        # authentication setting
        _auth_settings: list[str] = ["labid_token"]

        return self.api_client.param_serialize(
            method="GET",
            resource_path="/vardok-migration",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth,
        )

    @validate_call
    def get_vardok_vardef_mapping_by_id(
        self,
        id: Annotated[
            StrictStr,
            Field(description="The ID of the definition in Vardok or Vardef."),
        ],
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> GetVardokVardefMappingById200Response:
        """Get one variable definition by vardok id or get the vardok id by vardef id.

        Get one variable definition by vardok id or get the vardok id by vardef id.

        :param id: The ID of the definition in Vardok or Vardef. (required)
        :type id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._get_vardok_vardef_mapping_by_id_serialize(
            id=id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "200": "GetVardokVardefMappingById200Response",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def get_vardok_vardef_mapping_by_id_with_http_info(
        self,
        id: Annotated[
            StrictStr,
            Field(description="The ID of the definition in Vardok or Vardef."),
        ],
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[GetVardokVardefMappingById200Response]:
        """Get one variable definition by vardok id or get the vardok id by vardef id.

        Get one variable definition by vardok id or get the vardok id by vardef id.

        :param id: The ID of the definition in Vardok or Vardef. (required)
        :type id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._get_vardok_vardef_mapping_by_id_serialize(
            id=id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "200": "GetVardokVardefMappingById200Response",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def get_vardok_vardef_mapping_by_id_without_preload_content(
        self,
        id: Annotated[
            StrictStr,
            Field(description="The ID of the definition in Vardok or Vardef."),
        ],
        _request_timeout: None
        | Annotated[StrictFloat, Field(gt=0)]
        | tuple[
            Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]
        ] = None,
        _request_auth: dict[StrictStr, Any] | None = None,
        _content_type: StrictStr | None = None,
        _headers: dict[StrictStr, Any] | None = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Get one variable definition by vardok id or get the vardok id by vardef id.

        Get one variable definition by vardok id or get the vardok id by vardef id.

        :param id: The ID of the definition in Vardok or Vardef. (required)
        :type id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """
        _param = self._get_vardok_vardef_mapping_by_id_serialize(
            id=id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index,
        )

        _response_types_map: dict[str, str | None] = {
            "200": "GetVardokVardefMappingById200Response",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout
        )
        return response_data.response

    def _get_vardok_vardef_mapping_by_id_serialize(
        self,
        id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:
        _host = None

        _collection_formats: dict[str, str] = {}

        _path_params: dict[str, str] = {}
        _query_params: list[tuple[str, str]] = []
        _header_params: dict[str, str | None] = _headers or {}
        _form_params: list[tuple[str, str]] = []
        _files: dict[
            str, str | bytes | list[str] | list[bytes] | list[tuple[str, bytes]]
        ] = {}
        _body_params: bytes | None = None

        # process the path parameters
        if id is not None:
            _path_params["id"] = id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if "Accept" not in _header_params:
            _header_params["Accept"] = self.api_client.select_header_accept(
                ["application/json"]
            )

        # authentication setting
        _auth_settings: list[str] = ["labid_token"]

        return self.api_client.param_serialize(
            method="GET",
            resource_path="/vardok-migration/{id}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth,
        )



"""Document dataset."""

from datadoc_model.all_optional import model

from ._merge import InconsistentDatasetsWarning
from .core import Datadoc
from .dapla_dataset_path_info import DaplaDatasetPathInfo
from .model_validation import ObligatoryDatasetWarning
from .model_validation import ObligatoryVariableWarning
from .utility import enums



import logging
from datetime import date
from os import PathLike
from pathlib import Path

from dapla_metadata.variable_definitions._generated.vardef_client.api.data_migration_api import (
    DataMigrationApi,
)
from dapla_metadata.variable_definitions._generated.vardef_client.api.draft_variable_definitions_api import (
    DraftVariableDefinitionsApi,
)
from dapla_metadata.variable_definitions._generated.vardef_client.api.variable_definitions_api import (
    VariableDefinitionsApi,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.complete_response import (
    CompleteResponse,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.draft import (
    Draft,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.vardok_id_response import (
    VardokIdResponse,
)
from dapla_metadata.variable_definitions._utils._client import VardefClient
from dapla_metadata.variable_definitions._utils.template_files import (
    _find_latest_template_file,
)
from dapla_metadata.variable_definitions._utils.template_files import (
    create_template_yaml,
)
from dapla_metadata.variable_definitions._utils.variable_definition_files import (
    _read_file_to_model,
)
from dapla_metadata.variable_definitions.exceptions import VariableNotFoundError
from dapla_metadata.variable_definitions.exceptions import vardef_exception_handler
from dapla_metadata.variable_definitions.exceptions import vardef_file_error_handler
from dapla_metadata.variable_definitions.vardok_id import VardokId
from dapla_metadata.variable_definitions.vardok_vardef_id_pair import VardokVardefIdPair
from dapla_metadata.variable_definitions.variable_definition import VariableDefinition

logger = logging.getLogger(__name__)


class Vardef:
    """Create, maintain and read Variable Definitions.

    ====================
    Variable Definitions
    ====================

    Variable Definitions are centralized definitions of concrete variables which are typically present in multiple datasets. Variable Definitions
    support standardization of data and metadata and facilitate sharing and joining of data by clarifying when variables have an identical
    definition.

    The methods in this class allow for creation, maintenance and access of Variable Definitions.

    Creation and maintenance of variables may only be performed by Statistics Norway employees representing a specific Dapla team, who are defined
    as the owners of a given Variable Definition. The group a user represents is chosen when starting a service in Dapla Lab. This class will
    seamlessly detect this and send it to the API. All maintenance is to be performed by the owners, with no intervention from administrators.

    ======
    Status
    ======
    All Variable Definitions have an associated status. The possible values for status are `DRAFT`, `PUBLISHED_INTERNAL` and `PUBLISHED_EXTERNAL`.

    -----
    Draft
    -----
    When a Variable Definition is created it is assigned the status `DRAFT`. Under this status the Variable Definition is:
    * Only visible to Statistics Norway employees.
    * Mutable (it may be changed directly without need for versioning).
    * Not suitable to refer to from other systems.
    This status may be changed to `PUBLISHED_INTERNAL` or `PUBLISHED_EXTERNAL` with a direct update.

    ------------------
    Published Internal
    ------------------
    Under this status the Variable Definition is:
    * Only visible to Statistics Norway employees.
    * Immutable (all changes are versioned).
    * Suitable to refer to in internal systems for statistics production.
    * Not suitable to refer to for external use (for example in Statistikkbanken).

    This status may be changed to `PUBLISHED_EXTERNAL` by creating a Patch version.

    ------------------
    Published External
    ------------------
    Under this status the Variable Definition is:
    * Visible to the general public.
    * Immutable (all changes are versioned).
    * Suitable to refer to from any system.

    This status may not be changed as it would break immutability. If a Variable Definition is no longer relevant then its period of validity
    should be ended by specifying a `valid_until` date in a Patch version.

    ============
    Immutability
    ============
    Variable Definitions are immutable. This means that any changes must be performed in a strict versioning system. Consumers can avoid
    being exposed to breaking changes by specifying a `date_of_validity` when they request a Variable Definition.
    """

    @classmethod
    @vardef_exception_handler
    def create_draft(cls, draft: Draft) -> VariableDefinition:
        """Create a Draft Variable Definition."""
        new_variable = VariableDefinition.from_model(
            DraftVariableDefinitionsApi(
                VardefClient.get_client(),
            ).create_variable_definition(
                draft=draft,
            ),
        )

        logger.info(
            "✅ Successfully created variable definition '%s' with ID '%s'",
            new_variable.short_name,
            new_variable.id,
        )
        return new_variable

    @classmethod
    @vardef_file_error_handler
    def create_draft_from_file(
        cls,
        file_path: PathLike[str] | None = None,
    ) -> VariableDefinition:
        """Create a Draft Variable Definition from a stored yaml file.

        By default the latest template file in the default directory is chosen, this may be overridden by providing a value for the optional `file_path` parameter.

        Args:
            file_path (PathLike[str], optional): Supply a file path to override the automatic one. Defaults to None.

        Raises:
            FileNotFoundError: When a file can't be found.

        Returns:
            VariableDefinition: The created draft variable definition.
        """
        return cls.create_draft(
            _read_file_to_model(
                file_path or _find_latest_template_file(),
                Draft,
            ),
        )

    @classmethod
    @vardef_exception_handler
    def migrate_from_vardok(cls, vardok_id: str) -> VariableDefinition:
        """Migrate a Variable Definition from Vardok to Vardef.

        - Each Vardok Variable Definition may only be migrated once.
        - The Dapla team of the person who performs the migration will be set as the owner.
        - All metadata should be checked for correctness before publication.

        Args:
            vardok_id (str): The ID of a Variable Definition in Vardok.

        Returns:
            VariableDefinition: The migrated Variable Definition in Vardef.
        """
        migrated_variable = VariableDefinition.from_model(
            DataMigrationApi(
                VardefClient.get_client(),
            ).create_variable_definition_from_var_dok(
                vardok_id=vardok_id,
            ),
        )

        logger.info(
            "✅ Successfully migrated variable definition '%s' with ID '%s'",
            migrated_variable.short_name,
            migrated_variable.id,
        )
        return migrated_variable

    @classmethod
    @vardef_exception_handler
    def list_vardok_vardef_mapping(cls) -> list[VardokVardefIdPair]:
        """List the mapping between vardok and vardef.

        Returns:
            List[VardokVardefIdPair]: The list with mappings between Vardok and Vardef
        """
        return [
            VardokVardefIdPair.from_model(definition)
            for definition in DataMigrationApi(
                VardefClient.get_client(),
            ).get_vardok_vardef_mapping()
        ]

    @classmethod
    @vardef_exception_handler
    def get_variable_definition_by_vardok_id(
        cls,
        vardok_id: str,
    ) -> VariableDefinition:
        """Get a Variable Definition by its Vardok ID.

        Args:
            vardok_id (str): The Vardok ID of the desired Variable Definition

        Returns:
            VariableDefinition: The Variable Definition.

        Raises:
            TypeError: If the incorrect type is returned.
        """
        raw_response = DataMigrationApi(
            VardefClient.get_client()
        ).get_vardok_vardef_mapping_by_id(
            vardok_id,
        )

        if isinstance(raw_response.actual_instance, CompleteResponse):
            return VariableDefinition.from_model(raw_response.actual_instance)
        msg = "Unexpected response type"
        raise TypeError(msg)

    @classmethod
    @vardef_exception_handler
    def get_vardok_id_by_short_name(
        cls,
        short_name: str,
    ) -> VardokId:
        """Retrieve a Vardok id by short name.

        Args:
            short_name (str): The short name of the desired Variable Definition

        Raises:
            TypeError: If the incorrect type is returned.
        """
        variable_definition = cls.get_variable_definition_by_shortname(short_name)

        raw_response = DataMigrationApi(
            VardefClient.get_client()
        ).get_vardok_vardef_mapping_by_id(
            variable_definition.id,
        )

        if isinstance(raw_response.actual_instance, VardokIdResponse):
            return VardokId.from_model(raw_response.actual_instance)
        msg = "Unexpected response type"
        raise TypeError(msg)

    @classmethod
    @vardef_exception_handler
    def list_variable_definitions(
        cls,
        date_of_validity: date | None = None,
    ) -> list[VariableDefinition]:
        """List variable definitions.

        ---------
        Filtering
        ---------
        If no filter arguments are provided then all Variable Definitions are returned. See the documentation for the
        individual arguments to understand their effect. Filter arguments are combined with AND logic.

        Args:
            date_of_validity (date | None, optional): List only variable definitions which are valid on this date. Defaults to None.

        Returns:
            list[VariableDefinition]: The list of Variable Definitions.
        """
        return [
            VariableDefinition.from_model(definition)
            for definition in VariableDefinitionsApi(
                VardefClient.get_client(),
            ).list_variable_definitions(
                date_of_validity=date_of_validity,
            )
        ]

    @classmethod
    @vardef_exception_handler
    def get_variable_definition_by_id(
        cls,
        variable_definition_id: str,
        date_of_validity: date | None = None,
    ) -> VariableDefinition:
        """Get a Variable Definition by ID.

        Args:
            variable_definition_id (str): The ID of the desired Variable Definition
            date_of_validity (date | None, optional): List only variable definitions which are valid on this date. Defaults to None.

        Returns:
            VariableDefinition: The Variable Definition.

        Raises:
            NotFoundException when the given ID is not found
        """
        return VariableDefinition.from_model(
            VariableDefinitionsApi(
                VardefClient.get_client(),
            ).get_variable_definition_by_id(
                variable_definition_id=variable_definition_id,
                date_of_validity=date_of_validity,
            ),
        )

    @classmethod
    @vardef_exception_handler
    def get_variable_definition_by_shortname(
        cls,
        short_name: str,
        date_of_validity: date | None = None,
    ) -> VariableDefinition:
        """Retrieve a Variable Definition by short name.

        Args:
            short_name (str): The short name of the Variable Definition.
            date_of_validity (date | None, optional): Filter by validity date. Defaults to None.

        Returns:
            VariableDefinition: The retrieved Variable Definition.

        Raises:
            VariableNotFoundError: If no matching Variable Definition is found.
            ValueError: If multiple variables with the same shortname is found.
        """
        client = VardefClient.get_client()
        api = VariableDefinitionsApi(client)

        variable_definitions = api.list_variable_definitions(
            short_name=short_name,
            date_of_validity=date_of_validity,
        )

        if not variable_definitions:
            msg = f"Variable with short name {short_name} not found"
            raise VariableNotFoundError(msg)
        if len(variable_definitions) > 1:
            msg = f"Lookup by short name {short_name} found multiple variables which should not be possible."
            raise VariableNotFoundError(msg)

        return VariableDefinition.from_model(variable_definitions[0])

    @classmethod
    @vardef_file_error_handler
    def write_template_to_file(cls, custom_file_path: str | None = None) -> Path:
        """Write template with default values to a yaml file."""
        file_path = create_template_yaml(
            custom_directory=Path(custom_file_path) if custom_file_path else None,
        )
        logger.info(
            f"✅ Created editable variable definition template file at {file_path}",  # noqa: G004
        )
        return file_path

    @classmethod
    @vardef_exception_handler
    def does_short_name_exist(
        cls,
        short_name: str,
    ) -> bool:
        """Return True if the short name exists in Vardef, otherwise False."""
        variable_definitions = Vardef.list_variable_definitions()
        for variable in variable_definitions:
            if short_name.strip() == variable.short_name:
                logger.info(
                    f"Found duplicate short name {short_name}",  # noqa: G004
                )
                return True
        return False



"""Nox sessions."""

import os
import shlex
import shutil
import sys
from pathlib import Path
from textwrap import dedent

import nox

package = "dapla_metadata"
python_versions = ["3.10", "3.11", "3.12", "3.13", "3.14"]
nox.needs_version = ">= 2021.6.6"
nox.options.default_venv_backend = "uv"
nox.options.sessions = (
    "pre-commit",
    "mypy",
    "tests",
    "typeguard",
    "xdoctest",
    "docs-build",
)

TESTS_SESSIONS_DEPENDENCIES = {
    "coverage[toml]",
    "pytest",
    "pygments",
    "pytest-mock",
    "pytest-asyncio",
    "requests-mock",
    "faker",
    "testcontainers[generic]",
    "typeguard"
}


def activate_virtualenv_in_precommit_hooks(session: nox.Session) -> None:
    """Activate virtualenv in hooks installed by pre-commit.

    This function patches git hooks installed by pre-commit to activate the
    session's virtual environment. This allows pre-commit to locate hooks in
    that environment when invoked from git.

    Args:
        session: The Session object.
    """
    assert session.bin is not None  # nosec

    # Only patch hooks containing a reference to this session's bindir. Support
    # quoting rules for Python and bash, but strip the outermost quotes so we
    # can detect paths within the bindir, like <bindir>/python.
    bindirs = [
        bindir[1:-1] if bindir[0] in "'\"" else bindir
        for bindir in (repr(session.bin), shlex.quote(session.bin))
    ]

    virtualenv = session.env.get("VIRTUAL_ENV")
    if virtualenv is None:
        return

    headers = {
        # pre-commit < 2.16.0
        "python": f"""\
            import os
            os.environ["VIRTUAL_ENV"] = {virtualenv!r}
            os.environ["PATH"] = os.pathsep.join((
                {session.bin!r},
                os.environ.get("PATH", ""),
            ))
            """,
        # pre-commit >= 2.16.0
        "bash": f"""\
            VIRTUAL_ENV={shlex.quote(virtualenv)}
            PATH={shlex.quote(session.bin)}"{os.pathsep}$PATH"
            """,
        # pre-commit >= 2.17.0 on Windows forces sh shebang
        "/bin/sh": f"""\
            VIRTUAL_ENV={shlex.quote(virtualenv)}
            PATH={shlex.quote(session.bin)}"{os.pathsep}$PATH"
            """,
    }

    hookdir = Path(".git") / "hooks"
    if not hookdir.is_dir():
        return

    for hook in hookdir.iterdir():
        if hook.name.endswith(".sample") or not hook.is_file():
            continue

        if not hook.read_bytes().startswith(b"#!"):
            continue

        text = hook.read_text()

        if not is_bindir_in_text(bindirs, text):
            continue

        lines = text.splitlines()
        hook.write_text(insert_header_in_hook(headers, lines))


def is_bindir_in_text(bindirs: list[str], text: str) -> bool:
    """Helper function to check if bindir is in text."""
    return any(
        Path("A") == Path("a") and bindir.lower() in text.lower() or bindir in text
        for bindir in bindirs
    )


def insert_header_in_hook(header: dict[str, str], lines: list[str]) -> str:
    """Helper function to insert headers in hook's text."""
    for executable, header_text in header.items():
        if executable in lines[0].lower():
            lines.insert(1, dedent(header_text))
            return "\n".join(lines)
    return "\n".join(lines)


@nox.session(name="pre-commit", python=python_versions[-1])
def precommit(session: nox.Session) -> None:
    """Lint using pre-commit."""
    args = session.posargs or [
        "run",
        "--all-files",
        "--hook-stage=manual",
        "--show-diff-on-failure",
    ]
    session.install(
        "pre-commit",
        "pre-commit-hooks",
    )
    session.run("pre-commit", *args)
    if args and args[0] == "install":
        activate_virtualenv_in_precommit_hooks(session)


@nox.session(python=python_versions)
def mypy(session: nox.Session) -> None:
    """Type-check using mypy."""
    args = session.posargs or ["src", "tests"]
    session.install(".")
    session.install(
        "mypy",
        "pytest",
        "types-setuptools",
        "pandas-stubs",
        "pyarrow-stubs",
        "types-Pygments",
        "types-colorama",
        "types-requests",
        "types-beautifulsoup4",
        "types-docker",
        "faker",
        "typeguard"
    )
    session.run("mypy", *args)
    if not session.posargs:
        session.run("mypy", f"--python-executable={sys.executable}", "noxfile.py")


@nox.session(python=python_versions)
def tests(session: nox.Session) -> None:
    """Run the test suite."""
    session.install(".")
    session.install(*TESTS_SESSIONS_DEPENDENCIES)
    try:
        session.run(
            "coverage",
            "run",
            "--parallel",
            "-m",
            "pytest",
            "-o",
            "pythonpath=",
            *session.posargs,
        )
    finally:
        if session.interactive:
            session.notify("coverage", posargs=[])


@nox.session(python=python_versions[-1])
def coverage(session: nox.Session) -> None:
    """Produce the coverage report."""
    args = session.posargs or ["report", "--skip-empty"]

    session.install("coverage[toml]")

    if not session.posargs and any(Path().glob(".coverage.*")):
        session.run("coverage", "combine")

    session.run("coverage", *args)


@nox.session(python=python_versions[-1])
def typeguard(session: nox.Session) -> None:
    """Runtime type checking using Typeguard."""
    session.install(".")
    session.install(*TESTS_SESSIONS_DEPENDENCIES, "typeguard")
    session.run("pytest", f"--typeguard-packages={package}.datasets", *session.posargs)


@nox.session(python=python_versions[-1])
def xdoctest(session: nox.Session) -> None:
    """Run examples with xdoctest."""
    if session.posargs:
        args = [package, *session.posargs]
    else:
        args = [f"--modname={package}", "--command=all"]
        if "FORCE_COLOR" in os.environ:
            args.append("--colored=1")

    session.install(".")
    session.install("xdoctest[colors]")
    session.run("python", "-m", "xdoctest", *args)


@nox.session(name="docs-build", python=python_versions[-1])
def docs_build(session: nox.Session) -> None:
    """Build the documentation."""
    args = session.posargs or ["docs", "docs/_build"]
    if not session.posargs and "FORCE_COLOR" in os.environ:
        args.insert(0, "--color")

    session.install(".")
    session.install(
        "sphinx", "sphinx-autodoc-typehints", "sphinx-click", "furo", "myst-parser"
    )

    build_dir = Path("docs", "_build")
    if build_dir.exists():
        shutil.rmtree(build_dir)

    session.run("sphinx-build", *args)


@nox.session(python=python_versions[-1])
def docs(session: nox.Session) -> None:
    """Build and serve the documentation with live reloading on file changes."""
    args = session.posargs or ["--open-browser", "docs", "docs/_build"]
    session.install(".")
    session.install(
        "sphinx",
        "sphinx-autobuild",
        "sphinx-autodoc-typehints",
        "sphinx-click",
        "furo",
        "myst-parser",
    )

    build_dir = Path("docs", "_build")
    if build_dir.exists():
        shutil.rmtree(build_dir)

    session.run("sphinx-autobuild", *args)



"""Validate, parse and render URNs."""

import logging
import re
from collections.abc import Iterable
from dataclasses import dataclass
from enum import Enum
from enum import auto
from typing import Literal

from pydantic import AnyUrl

from dapla_metadata._shared.config import get_dapla_environment
from dapla_metadata._shared.enums import DaplaEnvironment
from dapla_metadata.datasets.utility.utils import VariableListType

logger = logging.getLogger(__name__)

URN_ERROR_MESSAGE_BASE = "The URL is not in a supported format"

URN_ERROR_MESSAGE_TEMPLATE = (
    URN_ERROR_MESSAGE_BASE
    + " for field '{field_name}' of variable '{short_name}'. URL: '{value}'. Please contact Team Metadata if this URL should be supported."
)


VARDEF_URL_TEMPLATE = "https://{subdomain}.{domain}/variable-definitions"


class SsbNaisDomains(str, Enum):
    """The available domains on SSBs Nais instance."""

    TEST_EXTERNAL = "test.ssb.no"
    TEST_INTERNAL = "intern.test.ssb.no"
    PROD_EXTERNAL = "ssb.no"
    PROD_INTERNAL = "intern.ssb.no"


class ReferenceUrlTypes(Enum):
    """The general category of the URL.

    This can be useful to refer to when constructing a URL from a URN for a
    specific context.
    """

    API = auto()
    FRONTEND = auto()


UrlVisibility = Literal["public", "internal"]


@dataclass
class UrnConverter:
    """Converts URLs to URNs and vice versa.

    Attributes:
        urn_base: The format for the URN, up to the identifier.
        id_pattern: A capturing group pattern which matches identifiers for this resource.
        url_bases: The list of all the different URL representations for a resource. There
            will typically be a number of URL representations for a particular resource,
            depending on which system or technology they are accessed through and other
            technical factors. This list defines which concrete URLs can be considered
            equivalent to a URN.
    """

    urn_base: str
    id_pattern: str
    url_bases: list[tuple[ReferenceUrlTypes, str]]

    def _extract_id(self, url: str, pattern: re.Pattern[str]) -> str | None:
        if match := pattern.match(url):
            return match.group(1)
        return None

    def _build_pattern(self, url_base: str) -> re.Pattern[str]:
        return re.compile(f"^{url_base}/{self.id_pattern}")

    def get_urn(self, identifier: str) -> str:
        """Build a URN for the given identifier."""
        return f"{self.urn_base}:{identifier}"

    def get_url(
        self,
        identifier: str,
        url_type: ReferenceUrlTypes,
        visibility: Literal["public", "internal"] = "public",
    ) -> str | None:
        """Build concrete URL to reference a resource.

        There are typically multiple URLs used to refer to one resource, this method attempts to support known variations.

        Args:
            identifier (str): The identifier of the resource the URL refers to.
            url_type (ReferenceUrlTypes): The representation type of the URL
            visibility (UrlVisibility, optional): Whether the URL should be that which is publicly available or not. Defaults to "public".

        Returns:
            str | None: The concrete URL. None if we cannot satisfy the supplied requirements.
        """
        candidates = [base[-1] for base in self.url_bases if base[0] == url_type]

        def matches_visibility(url: str, visibility: UrlVisibility):
            return (".intern." in url) is (visibility == "internal")

        def matches_environment(url: str):
            current_environment = get_dapla_environment()
            if current_environment == DaplaEnvironment.TEST:
                return ".test." in url
            return ".test." not in url

        if url := next(
            (
                url
                for url in candidates
                if matches_visibility(url, visibility) and matches_environment(url)
            ),
            None,
        ):
            return url + "/" + identifier
        return None

    def get_id(self, urn_or_url: str | AnyUrl) -> str | None:
        """Get an identifier from a URN or URL.

        Args:
            urn_or_url (str | AnyUrl): The URN or URL refering to a particular resource

        Returns:
            str | None: The identifier for the resource, or None if it cannot be extracted.
        """
        if str(urn_or_url).startswith(self.urn_base):
            return str(urn_or_url).removeprefix(self.urn_base + ":")
        return self._extract_id_from_url(urn_or_url)

    def is_id(self, value: str) -> bool:
        """Check if the value is an identifier for this URN type.

        Args:
            value (str): The value to check.
        """
        if not isinstance(value, str):
            # Mypy thinks it's impossible to reach this branch, but there are no guarantees in Python.
            return False  # type: ignore [unreachable]
        pattern = re.compile(f"^{self.id_pattern}$")
        return bool(pattern.match(value))

    def _extract_id_from_url(self, url: str | AnyUrl) -> str | None:
        patterns = (self._build_pattern(url[-1]) for url in self.url_bases)
        matches = (self._extract_id(str(url), p) for p in patterns)
        return next((m for m in matches if m), None)

    def convert_url_to_urn(self, url: str | AnyUrl) -> AnyUrl | None:
        """Convert a URL to a generalized URN for that same resource.

        Args:
            url (str | AnyUrl): The URL to convert.

        Returns:
            str | None: The URN or None if it can't be converted.
        """
        if str(url).startswith(self.urn_base):
            # In this case the value is already in the expected format and nothing needs to be done.
            return AnyUrl(url)
        if identifier := self._extract_id_from_url(url):
            return AnyUrl(self.get_urn(identifier))

        return None


vardef_urn_converter = UrnConverter(
    urn_base="urn:ssb:variable-definition:vardef",
    id_pattern=r"([a-zA-Z0-9-_]{8})",  # 8 character Nanoid with default alphabet
    url_bases=[
        *[
            (
                ReferenceUrlTypes.API,
                VARDEF_URL_TEMPLATE.format(
                    subdomain="metadata", domain=nais_domain.value
                ),
            )
            for nais_domain in SsbNaisDomains
        ],
        *[
            (
                ReferenceUrlTypes.FRONTEND,
                VARDEF_URL_TEMPLATE.format(
                    subdomain="catalog", domain=nais_domain.value
                ),
            )
            for nais_domain in SsbNaisDomains
        ],
    ],
)

klass_urn_converter = UrnConverter(
    urn_base="urn:ssb:classification:klass",
    id_pattern=r"([0-9]{1,5})",
    url_bases=[
        (ReferenceUrlTypes.FRONTEND, "https://www.ssb.no/klass/klassifikasjoner"),
        (ReferenceUrlTypes.FRONTEND, "https://www.ssb.no/en/klass/klassifikasjoner"),
        (ReferenceUrlTypes.API, "https://data.ssb.no/api/klass/v1/classifications"),
    ],
)


def convert_uris_to_urns(
    variables: VariableListType, field_name: str, converters: Iterable[UrnConverter]
) -> None:
    """Where URIs are recognized URLs, convert them to URNs.

    Where the value is not a known URL we preserve the value as it is and log an
    ERROR level message.

    Args:
        variables (VariableListType): The list of variables.
        field_name (str): The name of the field which has URLs to convert to URNs
        converters (Iterable[UrnConverter]): One or more converters which implement
            conversion of URLs into one specific URN format. These will typically be
            specific to an individual metadata reference system.
    """
    for v in variables:
        field = getattr(v, field_name, None)
        if field:
            if urn := next((c.convert_url_to_urn(field) for c in converters), None):
                setattr(v, field_name, urn)
            else:
                logger.error(
                    URN_ERROR_MESSAGE_TEMPLATE.format(
                        field_name=field_name,
                        short_name=v.short_name,
                        value=field,
                    )
                )



"""Model Backwards Compatibility.

This package contains code for upgrading existing metadata documents to the newest version of the model.
This is analogous to a Database Migration where the structure of the data has changed and we wish to
retain already persisted information.

"""

from ._utils import is_metadata_in_container_structure
from .model_backwards_compatibility import upgrade_metadata
